import sys

from gi.repository import Gtk

from pychess.ic import GAME_TYPES_BY_RATING_TYPE, TYPE_WILD, WildGameType
from pychess.System.ping import Pinger
from pychess.System.Log import log
from pychess.widgets import mainwindow


class UserInfoSection:
    def __init__(self, widgets, connection, host, lounge):
        self.widgets = widgets
        self.connection = connection
        self.host = host
        self.lounge = lounge
        self.pinger = None
        self.ping_label = None

        self.dock = self.widgets["fingerTableDock"]

        self.connection.fm.connect("fingeringFinished", self.onFinger)
        self.connection.fm.finger(self.connection.getUsername())
        self.connection.bm.connect(
            "curGameEnded",
            lambda *args: self.connection.fm.finger(self.connection.getUsername()),
        )

        self.widgets["usernameLabel"].set_markup(
            "<b>%s</b>" % self.connection.getUsername()
        )

    def _del(self):
        if self.pinger is not None:
            self.pinger.stop()

    def onFinger(self, fm, finger):
        # print(finger.getName(), self.connection.getUsername())
        my_finger = finger.getName().lower() == self.connection.getUsername().lower()
        if my_finger:
            self.widgets["usernameLabel"].set_markup("<b>%s</b>" % finger.getName())
        rows = 1
        if finger.getRatingsLen() > 0:
            rows += finger.getRatingsLen() + 1
        if finger.getEmail():
            rows += 1
        if finger.getCreated():
            rows += 1

        cols = 6 if my_finger else 9
        table = Gtk.Table(cols, rows)
        table.props.column_spacing = 12
        table.props.row_spacing = 4

        def label(value, xalign=0, is_error=False):
            if is_error:
                label = Gtk.Label()
                label.set_markup(
                    '<span size="larger" foreground="red">' + value + "</span>"
                )
            else:
                label = Gtk.Label(label=value)
            label.props.xalign = xalign
            return label

        row = 0

        ELO, DEVIATION, WINS, LOSSES, DRAWS, TOTAL, BESTELO, BESTTIME = range(8)
        if finger.getRatingsLen() > 0:
            if my_finger:
                headers = (_("Rating"), _("Win"), _("Draw"), _("Loss"))
            else:
                headers = (
                    _("Rating"),
                    _("Need") if self.connection.ICC else _("RD"),
                    _("Win"),
                    _("Draw"),
                    _("Loss"),
                    _("Best"),
                )
            for i, item in enumerate(headers):
                table.attach(label(item, xalign=1), i + 1, i + 2, 0, 1)
            row += 1
            for rating_type, rating in finger.getRatings().items():
                col = 0
                ratinglabel = label(
                    GAME_TYPES_BY_RATING_TYPE[rating_type].display_text + ":"
                )
                table.attach(ratinglabel, col, col + 1, row, row + 1)
                col += 1
                if rating_type is TYPE_WILD:
                    ratinglabel.set_tooltip_text(
                        _(
                            'On FICS, your "Wild" rating encompasses all of the \
                        following variants at all time controls:\n'
                        )
                        + ", ".join(
                            [gt.display_text for gt in WildGameType.instances()]
                        )
                    )
                table.attach(label(rating[ELO], xalign=1), col, col + 1, row, row + 1)
                col += 1
                if not my_finger:
                    table.attach(
                        label(rating[DEVIATION], xalign=1), col, col + 1, row, row + 1
                    )
                    col += 1
                table.attach(label(rating[WINS], xalign=1), col, col + 1, row, row + 1)
                col += 1
                table.attach(label(rating[DRAWS], xalign=1), col, col + 1, row, row + 1)
                col += 1
                table.attach(
                    label(rating[LOSSES], xalign=1), col, col + 1, row, row + 1
                )
                col += 1
                if not my_finger and len(rating) > BESTELO:
                    best = rating[BESTELO] if int(rating[BESTELO]) > 0 else ""
                    table.attach(label(best, xalign=1), col, col + 1, row, row + 1)
                    col += 1
                    table.attach(
                        label(rating[BESTTIME], xalign=1), col, col + 1, row, row + 1
                    )
                    col += 1
                row += 1

            table.attach(Gtk.HSeparator(), 0, cols, row, row + 1, ypadding=2)
            row += 1

        if finger.getSanctions() != "":
            table.attach(label(_("Sanctions") + ":", is_error=True), 0, 1, row, row + 1)
            table.attach(label(finger.getSanctions()), 1, cols, row, row + 1)
            row += 1

        if finger.getEmail():
            table.attach(label(_("Email") + ":"), 0, 1, row, row + 1)
            table.attach(label(finger.getEmail()), 1, cols, row, row + 1)
            row += 1

        player = self.connection.players.get(finger.getName())
        if not self.connection.ICC and not player.isGuest():
            table.attach(label(_("Games") + ":"), 0, 1, row, row + 1)
            llabel = Gtk.Label()
            llabel.props.xalign = 0
            link = (
                "http://ficsgames.org/cgi-bin/search.cgi?player=%s" % finger.getName()
            )
            llabel.set_markup(f'<a href="{link}">{link}</a>')
            table.attach(llabel, 1, cols, row, row + 1)
            row += 1

        if finger.getCreated():
            table.attach(label(_("Spent") + ":"), 0, 1, row, row + 1)
            string = "{} {}".format(finger.getTotalTimeOnline(), _("online in total"))
            table.attach(label(string), 1, cols, row, row + 1)
            row += 1

        # TODO: ping causes random crashes on Windows
        if my_finger and sys.platform != "win32":
            table.attach(label(_("Ping") + ":"), 0, 1, row, row + 1)
            if self.ping_label:
                if self.dock.get_children():
                    self.dock.get_children()[0].remove(self.ping_label)
            else:
                if self.connection.ICC:
                    self.ping_label = Gtk.Label(label="")  # TODO
                else:
                    self.ping_label = Gtk.Label(label=_("Connecting") + "...")
                self.ping_label.props.xalign = 0

            def callback(pinger, pingtime):
                log.debug(
                    f"'{str(self.pinger)}' '{str(pingtime)}'",
                    extra={"task": (self.connection.username, "UIS.oF.callback")},
                )
                if isinstance(pingtime, str):
                    self.ping_label.set_text(pingtime)
                elif pingtime == -1:
                    self.ping_label.set_text(_("Unknown"))
                else:
                    self.ping_label.set_text("%.0f ms" % pingtime)

            if (not self.pinger) and (not self.connection.ICC):
                self.pinger = Pinger(self.host)
                self.pinger.start()
                self.pinger.connect("received", callback)
                self.pinger.connect("error", callback)
            table.attach(self.ping_label, 1, cols, row, row + 1)
            row += 1

        if not my_finger:
            if self.lounge.finger_sent:
                dialog = Gtk.MessageDialog(
                    mainwindow(), type=Gtk.MessageType.INFO, buttons=Gtk.ButtonsType.OK
                )
                dialog.set_title(_("Finger"))
                dialog.set_markup("<b>%s</b>" % finger.getName())
                table.show_all()
                dialog.get_message_area().add(table)
                dialog.run()
                dialog.destroy()
            self.lounge.finger_sent = False
            return

        if not self.connection.isRegistred():
            vbox = Gtk.VBox()
            table.attach(vbox, 0, cols, row, row + 1)
            label0 = Gtk.Label()
            label0.props.xalign = 0
            label0.props.wrap = True
            label0.props.width_request = 300
            if self.connection.ICC:
                reg = "https://store.chessclub.com/customer/account/create/"
                txt = _(
                    "You are currently logged in as a guest but "
                    + "there is a completely free trial for 30 days, "
                    + "and beyond that, there is no charge and "
                    + "the account would remain active with the ability to play games. "
                    + "(With some restrictions. For example, no premium videos, "
                    + "some limitations in channels, and so on.) "
                    + "To register an account, go to "
                )
            else:
                reg = "http://www.freechess.org/Register/index.html"
                txt = _(
                    "You are currently logged in as a guest. "
                    + "A guest can't play rated games and therefore isn't "
                    + "able to play as many of the types of matches offered as "
                    + "a registered user. To register an account, go to "
                )

            label0.set_markup(f'{txt} <a href="{reg}">{reg}</a>.')
            vbox.add(label0)

        if self.dock.get_children():
            self.dock.remove(self.dock.get_children()[0])
        self.dock.add(table)
        self.dock.show_all()



import asyncio

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject

from pychess.Utils.const import (
    DRAW_OFFER,
    ABORT_OFFER,
    ADJOURN_OFFER,
    TAKEBACK_OFFER,
    PAUSE_OFFER,
    RESUME_OFFER,
    RESIGNATION,
    FLAG_CALL,
    SWITCH_OFFER,
    HURRY_ACTION,
    ACTION_ERROR_NOT_OUT_OF_TIME,
    ACTION_ERROR_CLOCK_NOT_STARTED,
    ACTION_ERROR_SWITCH_UNDERWAY,
    ACTION_ERROR_TOO_LARGE_UNDO,
    ACTION_ERROR_NONE_TO_WITHDRAW,
    ACTION_ERROR_NONE_TO_ACCEPT,
    LOCAL,
    RUNNING,
    CHAT_ACTION,
    ACTION_ERROR_NONE_TO_DECLINE,
)
from pychess.Utils.logic import validate
from pychess.Utils.Move import Move
from pychess.Utils.Offer import Offer
from pychess.System.Log import log
from pychess.System import conf

from pychess.Players.Player import Player, PlayerIsDead, TurnInterrupt, PassInterrupt
from pychess.widgets.InfoBar import InfoBarMessage, InfoBarMessageButton
from pychess.widgets import InfoBar

OFFER_MESSAGES = {
    DRAW_OFFER: (
        _("Your opponent has offered you a draw."),
        _(
            "Your opponent has offered you a draw. If you accept this offer, the game will end with a score of 1/2 - 1/2."
        ),
        False,
    ),
    ABORT_OFFER: (
        _("Your opponent wants to abort the game."),
        _(
            "Your opponent has asked that the game be aborted. If you accept this offer, the game will end with no rating change."
        ),
        False,
    ),
    ADJOURN_OFFER: (
        _("Your opponent wants to adjourn the game."),
        _(
            "Your opponent has asked that the game be adjourned. If you accept this offer, the game will be adjourned and you can resume it later (when your opponent is online and both players agree to resume)."
        ),
        False,
    ),
    TAKEBACK_OFFER: (
        _("Your opponent wants to undo %s move(s)."),
        _(
            "Your opponent has asked that the last %s move(s) be undone. If you accept this offer, the game will continue from the earlier position."
        ),
        True,
    ),
    PAUSE_OFFER: (
        _("Your opponent wants to pause the game."),
        _(
            "Your opponent has asked that the game be paused. If you accept this offer, the game clock will be paused until both players agree to resume the game."
        ),
        False,
    ),
    RESUME_OFFER: (
        _("Your opponent wants to resume the game."),
        _(
            "Your opponent has asked that the game be resumed. If you accept this offer, the game clock will continue from where it was paused."
        ),
        False,
    ),
}

ACTION_NAMES = {
    RESIGNATION: _("The resignation"),
    FLAG_CALL: _("The flag call"),
    DRAW_OFFER: _("The draw offer"),
    ABORT_OFFER: _("The abort offer"),
    ADJOURN_OFFER: _("The adjourn offer"),
    PAUSE_OFFER: _("The pause offer"),
    RESUME_OFFER: _("The resume offer"),
    SWITCH_OFFER: _("The offer to switch sides"),
    TAKEBACK_OFFER: _("The takeback offer"),
}

ACTION_ACTIONS = {
    RESIGNATION: _("resign"),
    FLAG_CALL: _("call your opponents flag"),
    DRAW_OFFER: _("offer a draw"),
    ABORT_OFFER: _("offer an abort"),
    ADJOURN_OFFER: _("offer to adjourn"),
    PAUSE_OFFER: _("offer a pause"),
    RESUME_OFFER: _("offer to resume"),
    SWITCH_OFFER: _("offer to switch sides"),
    TAKEBACK_OFFER: _("offer a takeback"),
    HURRY_ACTION: _("ask your opponent to move"),
}

ERROR_MESSAGES = {
    ACTION_ERROR_NOT_OUT_OF_TIME: _("Your opponent is not out of time."),
    ACTION_ERROR_CLOCK_NOT_STARTED: _("The clock hasn't been started yet."),
    ACTION_ERROR_SWITCH_UNDERWAY: _("You can't switch colors during the game."),
    ACTION_ERROR_TOO_LARGE_UNDO: _("You have tried to undo too many moves."),
}


class Human(Player):
    __type__ = LOCAL

    __gsignals__ = {
        "messageReceived": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self, gmwidg, color, name, ichandle=None, icrating=None):
        Player.__init__(self)

        self.defname = "Human"
        self.board = gmwidg.board
        self.gmwidg = gmwidg
        self.gamemodel = self.gmwidg.gamemodel
        self.move_queue = asyncio.Queue()
        self.color = color

        self.board_cids = [
            self.board.connect("piece_moved", self.piece_moved),
            self.board.connect("action", self.emit_action),
        ]
        self.setName(name)
        self.ichandle = ichandle
        self.icrating = icrating

        self.timemodel_cid = None
        if self.gamemodel.timed:
            self.timemodel_cid = self.gamemodel.timemodel.connect(
                "zero_reached", self.zero_reached
            )
        self.cid = self.gamemodel.connect_after(
            "game_terminated", self.on_game_terminated
        )

    def on_game_terminated(self, model):
        for cid in self.board_cids:
            self.board.disconnect(cid)
        if self.gamemodel.timed and self.timemodel_cid is not None:
            self.gamemodel.timemodel.disconnect(self.timemodel_cid)
        self.gamemodel.disconnect(self.cid)

    # Handle signals from the board

    def zero_reached(self, timemodel, color):
        if (
            conf.get("autoCallFlag")
            and self.gamemodel.status == RUNNING
            and timemodel.getPlayerTime(1 - self.color) <= 0
        ):
            log.info(
                "Automatically sending flag call on behalf of player %s." % self.name
            )
            self.emit("offer", Offer(FLAG_CALL))

    def piece_moved(self, board, move, color):
        if color != self.color:
            return
        self.move_queue.put_nowait(move)

    def emit_action(self, board, action, player, param):
        # If there are two or more tabs open, we have to ensure us that it is
        # us who are in the active tab, and not the others
        if not self.gmwidg.isInFront():
            return

        # If there are two human players, we have to ensure us that it was us
        # who did the action, and not the others
        if self.gamemodel.players[1 - self.color].__type__ == LOCAL:
            if action == HURRY_ACTION:
                if player.color == self.color:
                    return
            else:
                if player.color != self.color:
                    return
        log.debug(f"Human.emit_action: self.name={self.name}, action={action}")
        self.emit("offer", Offer(action, param=param))

    # Send the player move updates
    async def makeMove(self, board1, move, board2):
        log.debug(f"Human.makeMove: move={move}, board1={board1} board2={board2}")
        if (
            self.board.view.premove_piece
            and self.board.view.premove0
            and self.board.view.premove1
            and self.color == self.board.view.premove_piece.color
        ):
            if validate(
                board1,
                Move(
                    self.board.view.premove0,
                    self.board.view.premove1,
                    board1,
                    promotion=self.board.view.premove_promotion,
                ),
            ):
                log.debug(
                    "Human.makeMove: Setting move to premove {} {}".format(
                        self.board.view.premove0, self.board.view.premove1
                    )
                )
                self.board.emit_move_signal(
                    self.board.view.premove0,
                    self.board.view.premove1,
                    promotion=self.board.view.premove_promotion,
                )
            # reset premove
            self.board.view.setPremove(None, None, None, None)
        self.gmwidg.setLocked(False)

        item = await self.move_queue.get()
        self.gmwidg.setLocked(True)

        if item == "del":
            log.debug("Human.makeMove got: del")
            raise PlayerIsDead
        elif item == "int":
            log.debug("Human.makeMove got: int")
            raise TurnInterrupt
        elif item == "pass":
            log.debug("Human.makeMove got: pass")
            raise PassInterrupt
        return item

    # Ending the game

    def end(self, status, reason):
        log.debug("Human.end: %s" % self.name)
        self.move_queue.put_nowait("del")

    def kill(self, reason):
        for num in self.conid:
            if self.board.handler_is_connected(num):
                self.board.disconnect(num)
        self.move_queue.put_nowait("del")

    # Interacting with the player

    def hurry(self):
        title = _("Your opponent asks you to hurry!")
        text = _(
            "Generally this means nothing, as the game is time-based, but if you want to \
            please your opponent, perhaps you should get going."
        )
        content = InfoBar.get_message_content(title, text, Gtk.STOCK_DIALOG_INFO)

        def response_cb(infobar, response, message):
            message.dismiss()

        message = InfoBarMessage(Gtk.MessageType.INFO, content, response_cb)
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.gmwidg.showMessage(message)

    def pause(self):
        self.gmwidg.setLocked(True)

    def resume(self):
        log.debug("Human.resume: %s" % (self))
        if self.board.view.model.curplayer == self:
            self.gmwidg.setLocked(False)

    def playerUndoMoves(self, movecount, gamemodel):
        log.debug(
            "Human.playerUndoMoves:  movecount=%s self=%s gamemodel.curplayer=%s"
            % (movecount, self, gamemodel.curplayer)
        )
        self.gmwidg.clearMessages()

        # If the movecount is odd, the player has changed, and we have to interupt
        if movecount % 2 == 1 and gamemodel.curplayer != self:
            # If it is no longer us to move, we raise TurnInterruprt in order to
            # let GameModel continue the game.
            log.debug(
                "Human.playerUndoMoves: putting TurnInterrupt into self.move_queue %s"
                % self.name
            )
            self.move_queue.put_nowait("int")

        # If the movecount is even, we have to ensure the board is unlocked.
        # This is because it might have been locked by the game ending, but
        # perhaps we have now undone some moves, and it is no longer ended.
        elif movecount % 2 == 0 and gamemodel.curplayer == self:
            log.debug(
                "Human.playerUndoMoves: self=%s: calling gmwidg.setLocked" % (self)
            )
            self.gmwidg.setLocked(False)
            log.debug(
                "Human.playerUndoMoves: putting PassInterrupt into self.move_queue %s"
                % self.name
            )
            self.move_queue.put_nowait("pass")

    def putMessage(self, text):
        self.emit("messageReceived", text)

    def sendMessage(self, text):
        self.emit("offer", Offer(CHAT_ACTION, param=text))

    # Offer handling

    def offer(self, offer):
        log.debug(f"Human.offer: self={self} {offer}")
        assert offer.type in OFFER_MESSAGES

        if self.gamemodel.players[1 - self.color].__type__ is LOCAL:
            self.emit("accept", offer)
            return

        heading, text, takes_param = OFFER_MESSAGES[offer.type]
        if takes_param:
            heading = heading % offer.param
            text = text % offer.param

        def response_cb(infobar, response, message):
            if response == Gtk.ResponseType.ACCEPT:
                if offer.type == TAKEBACK_OFFER:
                    self.gamemodel.undoMoves(offer.param)
                self.emit("accept", offer)
            elif response == Gtk.ResponseType.NO:
                self.emit("decline", offer)
            message.dismiss()

        content = InfoBar.get_message_content(heading, text, Gtk.STOCK_DIALOG_QUESTION)
        message = InfoBarMessage(Gtk.MessageType.QUESTION, content, response_cb)
        message.add_button(InfoBarMessageButton(_("Accept"), Gtk.ResponseType.ACCEPT))
        message.add_button(InfoBarMessageButton(_("Decline"), Gtk.ResponseType.NO))
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.gmwidg.showMessage(message)

    def offerDeclined(self, offer):
        log.debug(f"Human.offerDeclined: self={self} {offer}")
        assert offer.type in ACTION_NAMES
        heading = _("%s was declined by your opponent") % ACTION_NAMES[offer.type]
        text = _("Resend %s?" % ACTION_NAMES[offer.type].lower())
        content = InfoBar.get_message_content(heading, text, Gtk.STOCK_DIALOG_INFO)

        def response_cb(infobar, response, message):
            if response == Gtk.ResponseType.ACCEPT:
                self.emit("offer", offer)
            message.dismiss()

        message = InfoBarMessage(Gtk.MessageType.INFO, content, response_cb)
        message.add_button(InfoBarMessageButton(_("Resend"), Gtk.ResponseType.ACCEPT))
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.gmwidg.replaceMessages(message)

    def offerWithdrawn(self, offer):
        log.debug(f"Human.offerWithdrawn: self={self} {offer}")
        assert offer.type in ACTION_NAMES
        heading = _("%s was withdrawn by your opponent") % ACTION_NAMES[offer.type]
        text = _("Your opponent seems to have changed their mind.")
        content = InfoBar.get_message_content(heading, text, Gtk.STOCK_DIALOG_INFO)

        def response_cb(infobar, response, message):
            message.dismiss()

        message = InfoBarMessage(Gtk.MessageType.INFO, content, response_cb)
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.gmwidg.showMessage(message)

    def offerError(self, offer, error):
        log.debug(f"Human.offerError: self={self} error={error} {offer}")
        assert offer.type in ACTION_NAMES
        actionName = ACTION_NAMES[offer.type]
        if error == ACTION_ERROR_NONE_TO_ACCEPT:
            heading = _("Unable to accept %s") % actionName.lower()
            text = _("Probably because it has been withdrawn.")
        elif (
            error == ACTION_ERROR_NONE_TO_DECLINE
            or error == ACTION_ERROR_NONE_TO_WITHDRAW
        ):
            # If the offer was not there, it has probably already been either
            # declined or withdrawn.
            return
        else:
            heading = _("%s returns an error") % actionName
            text = ERROR_MESSAGES[error]

        content = InfoBar.get_message_content(heading, text, Gtk.STOCK_DIALOG_WARNING)

        def response_cb(infobar, response, message):
            message.dismiss()

        message = InfoBarMessage(Gtk.MessageType.WARNING, content, response_cb)
        message.add_button(
            InfoBarMessageButton(Gtk.STOCK_CLOSE, Gtk.ResponseType.CANCEL)
        )
        self.gmwidg.showMessage(message)



from collections import namedtuple

# https://www.iso.org/obp/ui/#iso:pub:PUB500001:en

ISO3166 = namedtuple("ISO3166", "iso2, country")
ISO3166_LIST = [
    ISO3166("unknown", _("Unknown")),
    # Specific to pyChess: ISO3166("C", _("Computer")),
    ISO3166("ad", _("Andorra")),
    ISO3166("ae", _("United Arab Emirates")),
    ISO3166("af", _("Afghanistan")),
    ISO3166("ag", _("Antigua and Barbuda")),
    ISO3166("ai", _("Anguilla")),
    ISO3166("al", _("Albania")),
    ISO3166("am", _("Armenia")),
    # Discontinued: ISO3166("an", _("Netherlands Antilles")),
    ISO3166("ao", _("Angola")),
    ISO3166("aq", _("Antarctica")),
    ISO3166("ar", _("Argentina")),
    ISO3166("as", _("American Samoa")),
    ISO3166("at", _("Austria")),
    ISO3166("au", _("Australia")),
    ISO3166("aw", _("Aruba")),
    ISO3166("ax", _("Åland Islands")),
    ISO3166("az", _("Azerbaijan")),
    ISO3166("ba", _("Bosnia and Herzegovina")),
    ISO3166("bb", _("Barbados")),
    ISO3166("bd", _("Bangladesh")),
    ISO3166("be", _("Belgium")),
    ISO3166("bf", _("Burkina Faso")),
    ISO3166("bg", _("Bulgaria")),
    ISO3166("bh", _("Bahrain")),
    ISO3166("bi", _("Burundi")),
    ISO3166("bj", _("Benin")),
    ISO3166("bl", _("Saint Barthélemy")),
    ISO3166("bm", _("Bermuda")),
    ISO3166("bn", _("Brunei Darussalam")),
    ISO3166("bo", _("Bolivia (Plurinational State of)")),
    ISO3166("bq", _("Bonaire, Sint Eustatius and Saba")),
    ISO3166("br", _("Brazil")),
    ISO3166("bs", _("Bahamas")),
    ISO3166("bt", _("Bhutan")),
    ISO3166("bv", _("Bouvet Island")),
    ISO3166("bw", _("Botswana")),
    ISO3166("by", _("Belarus")),
    ISO3166("bz", _("Belize")),
    ISO3166("ca", _("Canada")),
    ISO3166("cc", _("Cocos (Keeling) Islands")),
    ISO3166("cd", _("Congo (the Democratic Republic of the)")),
    ISO3166("cf", _("Central African Republic")),
    ISO3166("cg", _("Congo")),
    ISO3166("ch", _("Switzerland")),
    ISO3166("ci", _("Côte d'Ivoire")),
    ISO3166("ck", _("Cook Islands")),
    ISO3166("cl", _("Chile")),
    ISO3166("cm", _("Cameroon")),
    ISO3166("cn", _("China")),
    ISO3166("co", _("Colombia")),
    ISO3166("cr", _("Costa Rica")),
    ISO3166("cu", _("Cuba")),
    ISO3166("cv", _("Cabo Verde")),
    ISO3166("cw", _("Curaçao")),
    ISO3166("cx", _("Christmas Island")),
    ISO3166("cy", _("Cyprus")),
    ISO3166("cz", _("Czechia")),
    ISO3166("de", _("Germany")),
    ISO3166("dj", _("Djibouti")),
    ISO3166("dk", _("Denmark")),
    ISO3166("dm", _("Dominica")),
    ISO3166("do", _("Dominican Republic")),
    ISO3166("dz", _("Algeria")),
    ISO3166("ec", _("Ecuador")),
    ISO3166("ee", _("Estonia")),
    ISO3166("eg", _("Egypt")),
    ISO3166("eh", _("Western Sahara")),
    ISO3166("er", _("Eritrea")),
    ISO3166("es", _("Spain")),
    ISO3166("et", _("Ethiopia")),
    ISO3166("fi", _("Finland")),
    ISO3166("fj", _("Fiji")),
    ISO3166("fk", _("Falkland Islands [Malvinas]")),
    ISO3166("fm", _("Micronesia (Federated States of)")),
    ISO3166("fo", _("Faroe Islands")),
    ISO3166("fr", _("France")),
    ISO3166("ga", _("Gabon")),
    ISO3166("gb", _("United Kingdom of Great Britain and Northern Ireland")),
    ISO3166("gd", _("Grenada")),
    ISO3166("ge", _("Georgia")),
    ISO3166("gf", _("French Guiana")),
    ISO3166("gg", _("Guernsey")),
    ISO3166("gh", _("Ghana")),
    ISO3166("gi", _("Gibraltar")),
    ISO3166("gl", _("Greenland")),
    ISO3166("gm", _("Gambia")),
    ISO3166("gn", _("Guinea")),
    ISO3166("gp", _("Guadeloupe")),
    ISO3166("gq", _("Equatorial Guinea")),
    ISO3166("gr", _("Greece")),
    ISO3166("gs", _("South Georgia and the South Sandwich Islands")),
    ISO3166("gt", _("Guatemala")),
    ISO3166("gu", _("Guam")),
    ISO3166("gw", _("Guinea-Bissau")),
    ISO3166("gy", _("Guyana")),
    ISO3166("hk", _("Hong Kong")),
    ISO3166("hm", _("Heard Island and McDonald Islands")),
    ISO3166("hn", _("Honduras")),
    ISO3166("hr", _("Croatia")),
    ISO3166("ht", _("Haiti")),
    ISO3166("hu", _("Hungary")),
    ISO3166("id", _("Indonesia")),
    ISO3166("ie", _("Ireland")),
    ISO3166("il", _("Israel")),
    ISO3166("im", _("Isle of Man")),
    ISO3166("in", _("India")),
    ISO3166("io", _("British Indian Ocean Territory")),
    ISO3166("iq", _("Iraq")),
    ISO3166("ir", _("Iran (Islamic Republic of)")),
    ISO3166("is", _("Iceland")),
    ISO3166("it", _("Italy")),
    ISO3166("je", _("Jersey")),
    ISO3166("jm", _("Jamaica")),
    ISO3166("jo", _("Jordan")),
    ISO3166("jp", _("Japan")),
    ISO3166("ke", _("Kenya")),
    ISO3166("kg", _("Kyrgyzstan")),
    ISO3166("kh", _("Cambodia")),
    ISO3166("ki", _("Kiribati")),
    ISO3166("km", _("Comoros")),
    ISO3166("kn", _("Saint Kitts and Nevis")),
    ISO3166("kp", _("Korea (the Democratic People's Republic of)")),
    ISO3166("kr", _("Korea (the Republic of)")),
    ISO3166("kw", _("Kuwait")),
    ISO3166("ky", _("Cayman Islands")),
    ISO3166("kz", _("Kazakhstan")),
    ISO3166("la", _("Lao People's Democratic Republic")),
    ISO3166("lb", _("Lebanon")),
    ISO3166("lc", _("Saint Lucia")),
    ISO3166("li", _("Liechtenstein")),
    ISO3166("lk", _("Sri Lanka")),
    ISO3166("lr", _("Liberia")),
    ISO3166("ls", _("Lesotho")),
    ISO3166("lt", _("Lithuania")),
    ISO3166("lu", _("Luxembourg")),
    ISO3166("lv", _("Latvia")),
    ISO3166("ly", _("Libya")),
    ISO3166("ma", _("Morocco")),
    ISO3166("mc", _("Monaco")),
    ISO3166("md", _("Moldova (the Republic of)")),
    ISO3166("me", _("Montenegro")),
    ISO3166("mf", _("Saint Martin (French part)")),
    ISO3166("mg", _("Madagascar")),
    ISO3166("mh", _("Marshall Islands")),
    ISO3166("mk", _("Macedonia (the former Yugoslav Republic of)")),
    ISO3166("ml", _("Mali")),
    ISO3166("mm", _("Myanmar")),
    ISO3166("mn", _("Mongolia")),
    ISO3166("mo", _("Macao")),
    ISO3166("mp", _("Northern Mariana Islands")),
    ISO3166("mq", _("Martinique")),
    ISO3166("mr", _("Mauritania")),
    ISO3166("ms", _("Montserrat")),
    ISO3166("mt", _("Malta")),
    ISO3166("mu", _("Mauritius")),
    ISO3166("mv", _("Maldives")),
    ISO3166("mw", _("Malawi")),
    ISO3166("mx", _("Mexico")),
    ISO3166("my", _("Malaysia")),
    ISO3166("mz", _("Mozambique")),
    ISO3166("na", _("Namibia")),
    ISO3166("nc", _("New Caledonia")),
    ISO3166("ne", _("Niger")),
    ISO3166("nf", _("Norfolk Island")),
    ISO3166("ng", _("Nigeria")),
    ISO3166("ni", _("Nicaragua")),
    ISO3166("nl", _("Netherlands")),
    ISO3166("no", _("Norway")),
    ISO3166("np", _("Nepal")),
    ISO3166("nr", _("Nauru")),
    ISO3166("nu", _("Niue")),
    ISO3166("nz", _("New Zealand")),
    ISO3166("om", _("Oman")),
    ISO3166("pa", _("Panama")),
    ISO3166("pe", _("Peru")),
    ISO3166("pf", _("French Polynesia")),
    ISO3166("pg", _("Papua New Guinea")),
    ISO3166("ph", _("Philippines")),
    ISO3166("pk", _("Pakistan")),
    ISO3166("pl", _("Poland")),
    ISO3166("pm", _("Saint Pierre and Miquelon")),
    ISO3166("pn", _("Pitcairn")),
    ISO3166("pr", _("Puerto Rico")),
    ISO3166("ps", _("Palestine, State of")),
    ISO3166("pt", _("Portugal")),
    ISO3166("pw", _("Palau")),
    ISO3166("py", _("Paraguay")),
    ISO3166("qa", _("Qatar")),
    ISO3166("re", _("Réunion")),
    ISO3166("ro", _("Romania")),
    ISO3166("rs", _("Serbia")),
    ISO3166("ru", _("Russian Federation")),
    ISO3166("rw", _("Rwanda")),
    ISO3166("sa", _("Saudi Arabia")),
    ISO3166("sb", _("Solomon Islands")),
    ISO3166("sc", _("Seychelles")),
    ISO3166("sd", _("Sudan")),
    ISO3166("se", _("Sweden")),
    ISO3166("sg", _("Singapore")),
    ISO3166("sh", _("Saint Helena, Ascension and Tristan da Cunha")),
    ISO3166("si", _("Slovenia")),
    ISO3166("sj", _("Svalbard and Jan Mayen")),
    ISO3166("sk", _("Slovakia")),
    ISO3166("sl", _("Sierra Leone")),
    ISO3166("sm", _("San Marino")),
    ISO3166("sn", _("Senegal")),
    ISO3166("so", _("Somalia")),
    ISO3166("sr", _("Suriname")),
    ISO3166("ss", _("South Sudan")),
    ISO3166("st", _("Sao Tome and Principe")),
    ISO3166("sv", _("El Salvador")),
    ISO3166("sx", _("Sint Maarten (Dutch part)")),
    ISO3166("sy", _("Syrian Arab Republic")),
    ISO3166("sz", _("Eswatini")),
    ISO3166("tc", _("Turks and Caicos Islands")),
    ISO3166("td", _("Chad")),
    ISO3166("tf", _("French Southern Territories")),
    ISO3166("tg", _("Togo")),
    ISO3166("th", _("Thailand")),
    ISO3166("tj", _("Tajikistan")),
    ISO3166("tk", _("Tokelau")),
    ISO3166("tl", _("Timor-Leste")),
    ISO3166("tm", _("Turkmenistan")),
    ISO3166("tn", _("Tunisia")),
    ISO3166("to", _("Tonga")),
    # Discontinued: ISO3166("tp", _("East Timor")),
    ISO3166("tr", _("Turkey")),
    ISO3166("tt", _("Trinidad and Tobago")),
    ISO3166("tv", _("Tuvalu")),
    ISO3166("tw", _("Taiwan (Province of China)")),
    ISO3166("tz", _("Tanzania, United Republic of")),
    ISO3166("ua", _("Ukraine")),
    ISO3166("ug", _("Uganda")),
    ISO3166("um", _("United States Minor Outlying Islands")),
    ISO3166("us", _("United States of America")),
    ISO3166("uy", _("Uruguay")),
    ISO3166("uz", _("Uzbekistan")),
    ISO3166("va", _("Holy See")),
    ISO3166("vc", _("Saint Vincent and the Grenadines")),
    ISO3166("ve", _("Venezuela (Bolivarian Republic of)")),
    ISO3166("vg", _("Virgin Islands (British)")),
    ISO3166("vi", _("Virgin Islands (U.S.)")),
    ISO3166("vn", _("Viet Nam")),
    ISO3166("vu", _("Vanuatu")),
    ISO3166("wf", _("Wallis and Futuna")),
    ISO3166("ws", _("Samoa")),
    ISO3166("ye", _("Yemen")),
    ISO3166("yt", _("Mayotte")),
    # Discontinued: ISO3166("yu", _("Yugoslavia")),
    ISO3166("za", _("South Africa")),
    ISO3166("zm", _("Zambia")),
    ISO3166("zw", _("Zimbabwe")),
]

# Bubble sort for the translated countries
for i in range(len(ISO3166_LIST) - 1, 1, -1):
    for j in range(1, i - 1):
        if ISO3166_LIST[i].country < ISO3166_LIST[j].country:
            tmp = ISO3166_LIST[i]
            ISO3166_LIST[i] = ISO3166_LIST[j]
            ISO3166_LIST[j] = tmp



import re
import asyncio

from gi.repository import GObject

from pychess.System.Log import log
from pychess.Savers.pgn import msToClockTimeTag

from pychess.Utils.const import (
    WHITEWON,
    WON_RESIGN,
    WON_DISCONNECTION,
    WON_CALLFLAG,
    BLACKWON,
    WON_MATE,
    WON_ADJUDICATION,
    WON_KINGEXPLODE,
    WON_NOMATERIAL,
    UNKNOWN_REASON,
    DRAW,
    DRAW_REPETITION,
    DRAW_BLACKINSUFFICIENTANDWHITETIME,
    DRAW_WHITEINSUFFICIENTANDBLACKTIME,
    DRAW_INSUFFICIENT,
    DRAW_CALLFLAG,
    DRAW_AGREE,
    DRAW_STALEMATE,
    DRAW_50MOVES,
    DRAW_LENGTH,
    DRAW_ADJUDICATION,
    ADJOURNED,
    ADJOURNED_COURTESY_WHITE,
    ADJOURNED_COURTESY_BLACK,
    ADJOURNED_COURTESY,
    ADJOURNED_AGREEMENT,
    ADJOURNED_LOST_CONNECTION_WHITE,
    ADJOURNED_LOST_CONNECTION_BLACK,
    ADJOURNED_LOST_CONNECTION,
    ADJOURNED_SERVER_SHUTDOWN,
    ABORTED,
    ABORTED_AGREEMENT,
    ABORTED_DISCONNECTION,
    ABORTED_EARLY,
    ABORTED_SERVER_SHUTDOWN,
    ABORTED_ADJUDICATION,
    ABORTED_COURTESY,
    UNKNOWN_STATE,
    BLACK,
    WHITE,
    reprFile,
    FISCHERRANDOMCHESS,
    CRAZYHOUSECHESS,
    WILDCASTLECHESS,
    WILDCASTLESHUFFLECHESS,
    ATOMICCHESS,
    LOSERSCHESS,
    SUICIDECHESS,
    GIVEAWAYCHESS,
    reprResult,
)

from pychess.ic import (
    IC_POS_OBSERVING_EXAMINATION,
    IC_POS_EXAMINATING,
    GAME_TYPES,
    IC_STATUS_PLAYING,
    BLKCMD_SEEK,
    BLKCMD_OBSERVE,
    BLKCMD_MATCH,
    TYPE_WILD,
    BLKCMD_SMOVES,
    BLKCMD_UNOBSERVE,
    BLKCMD_MOVES,
    BLKCMD_FLAG,
    parseRating,
)

from pychess.ic.FICSObjects import (
    FICSGame,
    FICSBoard,
    FICSHistoryGame,
    FICSAdjournedGame,
    FICSJournalGame,
    FICSPlayer,
)

names = r"(\w+)"
titles = r"((?:\((?:GM|IM|FM|WGM|WIM|WFM|TM|SR|TD|SR|CA|C|U|D|B|T|\*)\))+)?"
ratedexp = "(rated|unrated)"
ratings = r"\(\s*([0-9\ \-\+]{1,4}[P E]?|UNR)\)"

weekdays = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
months = [
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec",
]

# "Thu Oct 14, 20:36 PDT 2010"
dates = r"({})\s+({})\s+(\d+),\s+(\d+):(\d+)\s+([A-Z\?]+)\s+(\d{{4}})".format(
    "|".join(weekdays),
    "|".join(months),
)

# "2010-10-14 20:36 UTC"
datesFatICS = r"(\d{4})-(\d{2})-(\d{2})\s+(\d{2}):(\d{2})\s+(UTC)"

moveListHeader1Str = "{} {} vs. {} {} --- (?:{}|{})".format(
    names, ratings, names, ratings, dates, datesFatICS
)
moveListHeader1 = re.compile(moveListHeader1Str)
moveListHeader2Str = (
    r"%s ([^ ]+) match, initial time: (\d+) minutes, increment: (\d+) seconds\."
    % ratedexp
)
moveListHeader2 = re.compile(moveListHeader2Str, re.IGNORECASE)
sanmove = "([a-hx@OoPKQRBN0-8+#=-]{2,7})"
movetime = r"\((\d:)?(\d{1,2}):(\d\d)(?:\.(\d{1,3}))?\)"
moveListMoves = re.compile(
    r"\s*(\d+)\. +(?:%s|\.\.\.) +%s *(?:%s +%s)?"
    % (sanmove, movetime, sanmove, movetime)
)

creating0 = re.compile(
    r"Creating: %s %s %s %s %s ([^ ]+) (\d+) (\d+)(?: \(adjourned\))?"
    % (names, ratings, names, ratings, ratedexp)
)
creating1 = re.compile(
    r"{Game (\d+) \(%s vs\. %s\) (?:Creating|Continuing) %s ([^ ]+) match\."
    % (names, names, ratedexp)
)
pr = re.compile(r"<pr> ([\d ]+)")
sr = re.compile(r"<sr> ([\d ]+)")

fileToEpcord = (
    ("a3", "b3", "c3", "d3", "e3", "f3", "g3", "h3"),
    ("a6", "b6", "c6", "d6", "e6", "f6", "g6", "h6"),
)


def parse_reason(result, reason, wname=None):
    """
    Parse the result value and reason line string for the reason and return
    the result and reason the game ended.

    result -- The result of the game, if known. It can be "None", but if it
    is "DRAW", then wname must be supplied
    """
    if result in (WHITEWON, BLACKWON):
        if "resigns" in reason:
            reason = WON_RESIGN
        elif "disconnection" in reason:
            reason = WON_DISCONNECTION
        elif "time" in reason:
            reason = WON_CALLFLAG
        elif "checkmated" in reason:
            reason = WON_MATE
        elif "adjudication" in reason:
            reason = WON_ADJUDICATION
        elif "exploded" in reason:
            reason = WON_KINGEXPLODE
        elif "material" in reason:
            reason = WON_NOMATERIAL
        else:
            reason = UNKNOWN_REASON
    elif result == DRAW:
        assert wname is not None
        if "repetition" in reason:
            reason = DRAW_REPETITION
        elif "material" in reason and "time" in reason:
            if wname + " ran out of time" in reason:
                reason = DRAW_BLACKINSUFFICIENTANDWHITETIME
            else:
                reason = DRAW_WHITEINSUFFICIENTANDBLACKTIME
        elif "material" in reason:
            reason = DRAW_INSUFFICIENT
        elif "time" in reason:
            reason = DRAW_CALLFLAG
        elif "agreement" in reason:
            reason = DRAW_AGREE
        elif "stalemate" in reason:
            reason = DRAW_STALEMATE
        elif "50" in reason:
            reason = DRAW_50MOVES
        elif "length" in reason:
            # FICS has a max game length on 800 moves
            reason = DRAW_LENGTH
        elif "adjudication" in reason:
            reason = DRAW_ADJUDICATION
        else:
            reason = UNKNOWN_REASON
    elif result == ADJOURNED or "adjourned" in reason:
        result = ADJOURNED
        if "courtesy" in reason:
            if wname:
                if wname in reason:
                    reason = ADJOURNED_COURTESY_WHITE
                else:
                    reason = ADJOURNED_COURTESY_BLACK
            elif "white" in reason:
                reason = ADJOURNED_COURTESY_WHITE
            elif "black" in reason:
                reason = ADJOURNED_COURTESY_BLACK
            else:
                reason = ADJOURNED_COURTESY
        elif "agreement" in reason:
            reason = ADJOURNED_AGREEMENT
        elif "connection" in reason:
            if "white" in reason:
                reason = ADJOURNED_LOST_CONNECTION_WHITE
            elif "black" in reason:
                reason = ADJOURNED_LOST_CONNECTION_BLACK
            else:
                reason = ADJOURNED_LOST_CONNECTION
        elif "server" in reason:
            reason = ADJOURNED_SERVER_SHUTDOWN
        else:
            reason = UNKNOWN_REASON
    elif "aborted" in reason:
        result = ABORTED
        if "agreement" in reason:
            reason = ABORTED_AGREEMENT
        elif "moves" in reason:
            # lost connection and too few moves; game aborted *
            reason = ABORTED_DISCONNECTION
        elif "move" in reason:
            # Game aborted on move 1 *
            reason = ABORTED_EARLY
        elif "shutdown" in reason:
            reason = ABORTED_SERVER_SHUTDOWN
        elif "adjudication" in reason:
            reason = ABORTED_ADJUDICATION
        else:
            reason = UNKNOWN_REASON
    elif "courtesyadjourned" in reason:
        result = ADJOURNED
        reason = ADJOURNED_COURTESY
    elif "courtesyaborted" in reason:
        result = ABORTED
        reason = ABORTED_COURTESY
    else:
        result = UNKNOWN_STATE
        reason = UNKNOWN_REASON

    return result, reason


class BoardManager(GObject.GObject):
    __gsignals__ = {
        "playGameCreated": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "obsGameCreated": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "exGameCreated": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "exGameReset": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "gameUndoing": (GObject.SignalFlags.RUN_FIRST, None, (int, int)),
        "archiveGamePreview": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "boardSetup": (GObject.SignalFlags.RUN_FIRST, None, (int, str, str, str)),
        "timesUpdate": (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (
                int,
                int,
                int,
            ),
        ),
        "obsGameEnded": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "curGameEnded": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "obsGameUnobserved": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "madeExamined": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "madeUnExamined": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "gamePaused": (GObject.SignalFlags.RUN_FIRST, None, (int, bool)),
        "tooManySeeks": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "nonoWhileExamine": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "matchDeclined": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "player_on_censor": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "player_on_noplay": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "player_lagged": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "opp_not_out_of_time": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "req_not_fit_formula": (GObject.SignalFlags.RUN_FIRST, None, (object, str)),
    }

    castleSigns = {}
    queuedStyle12s = {}

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection
        self.connection.expect_line(self.onStyle12, "<12> (.+)")
        self.connection.expect_line(
            self.onWasPrivate, r"Sorry, game (\d+) is a private game\."
        )
        self.connection.expect_line(
            self.tooManySeeks, "You can only have 3 active seeks."
        )
        self.connection.expect_line(
            self.nonoWhileExamine,
            "(?:You cannot challenge while you are examining a game.)|"
            + "(?:You are already examining a game.)",
        )
        self.connection.expect_line(
            self.matchDeclined, "%s declines the match offer." % names
        )
        self.connection.expect_line(
            self.player_on_censor, "%s is censoring you." % names
        )
        self.connection.expect_line(
            self.player_on_noplay, "You are on %s's noplay list." % names
        )
        self.connection.expect_line(
            self.player_lagged, r"Game (\d+): %s has lagged for (\d+) seconds\." % names
        )
        self.connection.expect_line(
            self.opp_not_out_of_time,
            r"Opponent is not out of time, wait for server autoflag\.",
        )

        self.connection.expect_n_lines(
            self.req_not_fit_formula,
            "Match request does not fit formula for %s:" % names,
            "%s's formula: (.+)" % names,
        )

        self.connection.expect_line(
            self.on_game_remove,
            r"\{Game (\d+) \(([A-Za-z]+) vs\. ([A-Za-z]+)\) ([A-Za-z']+ .+)\} (\*|1/2-1/2|1-0|0-1)$",
        )

        if self.connection.USCN:
            self.connection.expect_n_lines(
                self.onPlayGameCreated,
                r"Creating: %s %s %s %s %s ([^ ]+) (\d+) (\d+)(?: \(adjourned\))?"
                % (names, ratings, names, ratings, ratedexp),
                "",
                r"{Game (\d+) \(%s vs\. %s\) (?:Creating|Continuing) %s ([^ ]+) match\."
                % (names, names, ratedexp),
                "",
                "<12> (.+)",
            )
        else:
            self.connection.expect_n_lines(
                self.onPlayGameCreated,
                r"Creating: %s %s %s %s %s ([^ ]+) (\d+) (\d+)(?: \(adjourned\))?"
                % (names, ratings, names, ratings, ratedexp),
                r"{Game (\d+) \(%s vs\. %s\) (?:Creating|Continuing) %s ([^ ]+) match\."
                % (names, names, ratedexp),
                "",
                "<12> (.+)",
            )

        # TODO: Trying to precisely match every type of possible response FICS
        # will throw at us for "Your seek matches..." or "Your seek qualifies
        # for [player]'s getgame" is error prone and we can never be sure we
        # even have all of the different types of replies the server will throw
        # at us. So we should probably make it possible for multi-line
        # prediction callbacks in VerboseTelnet to put lines the callback isn't
        # interested in or doesn't handle back onto the input line stack in
        # VerboseTelnet.TelnetLines
        self.connection.expect_fromto(
            self.onMatchingSeekOrGetGame,
            r"Your seek (?:matches one already posted by %s|qualifies for %s's getgame)\."
            % (names, names),
            "(?:<12>|<sn>) (.+)",
        )
        self.connection.expect_fromto(
            self.onInterceptedChallenge,
            r"Your challenge intercepts %s's challenge\." % names,
            "<12> (.+)",
        )

        if self.connection.USCN:
            self.connection.expect_n_lines(
                self.onObserveGameCreated,
                r"You are now observing game \d+\.",
                "",
                "<12> (.+)",
            )
        else:
            self.connection.expect_n_lines(
                self.onObserveGameCreated,
                r"You are now observing game \d+\.",
                r"Game (\d+): %s %s %s %s %s ([\w/]+) (\d+) (\d+)"
                % (names, ratings, names, ratings, ratedexp),
                "",
                "<12> (.+)",
            )

        self.connection.expect_fromto(
            self.onObserveGameMovesReceived,
            r"Movelist for game (\d+):",
            r"{Still in progress} \*",
        )

        self.connection.expect_fromto(
            self.onArchiveGameSMovesReceived,
            moveListHeader1Str,
            #                                       "\s*{((?:Game courtesyadjourned by (Black|White))|(?:Still in progress)|(?:Game adjourned by mutual agreement)|(?:(White|Black) lost connection; game adjourned)|(?:Game adjourned by ((?:server shutdown)|(?:adjudication)|(?:simul holder))))} \*")
            r"\s*{.*(?:([Gg]ame.*adjourned.\s*)|(?:Still in progress)|(?:Neither.*)|(?:Game drawn.*)|(?:White.*)|(?:Black.*)).*}\s*(?:(?:1/2-1/2)|(?:1-0)|(?:0-1))?\s*",
        )

        self.connection.expect_line(
            self.onGamePause, r"Game (\d+): Game clock (paused|resumed)\."
        )
        self.connection.expect_line(
            self.onUnobserveGame, r"Removing game (\d+) from observation list\."
        )

        self.connection.expect_line(
            self.made_examined, r"%s has made you an examiner of game (\d+)\." % names
        )

        self.connection.expect_line(
            self.made_unexamined, r"You are no longer examining game (\d+)\."
        )

        self.connection.expect_n_lines(
            self.onExamineGameCreated,
            r"Starting a game in examine \(scratch\) mode\.",
            "",
            "<12> (.+)",
        )

        self.queuedEmits = {}
        self.gamemodelStartedEvents = {}
        self.theGameImPlaying = None
        self.gamesImObserving = {}

        # The ms ivar makes the remaining second fields in style12 use ms
        self.connection.client.run_command("iset ms 1")
        # Style12 is a must, when you don't want to parse visualoptimized stuff
        self.connection.client.run_command("set style 12")
        # When we observe fischer games, this puts a startpos in the movelist
        self.connection.client.run_command("iset startpos 1")
        # movecase ensures that bc3 will never be a bishop move
        self.connection.client.run_command("iset movecase 1")
        # don't unobserve games when we start a new game
        self.connection.client.run_command("set unobserve 3")
        self.connection.lvm.autoFlagNotify()

        # gameinfo <g1> doesn't really have any interesting info, at least not
        # until we implement crasyhouse and stuff
        # self.connection.client.run_command("iset gameinfo 1")

    def start(self):
        self.connection.games.connect("FICSGameEnded", self.onGameEnd)

    @classmethod
    def parseStyle12(cls, line, castleSigns=None):
        # <12> rnbqkb-r pppppppp -----n-- -------- ----P--- -------- PPPPKPPP RNBQ-BNR
        # B -1 0 0 1 1 0 7 Newton Einstein 1 2 12 39 39 119 122 2 K/e1-e2 (0:06) Ke2 0
        fields = line.split()

        curcol = fields[8] == "B" and BLACK or WHITE
        gameno = int(fields[15])
        relation = int(fields[18])
        lastmove = fields[28] != "none" and fields[28] or None
        if lastmove is None:
            ply = 0
        else:
            ply = int(fields[25]) * 2 - (curcol == WHITE and 2 or 1)
        wname = fields[16]
        bname = fields[17]
        wms = int(fields[23])
        bms = int(fields[24])
        gain = int(fields[20])

        # Board data
        fenrows = []
        for row in fields[:8]:
            fenrow = []
            spaceCounter = 0
            for char in row:
                if char == "-":
                    spaceCounter += 1
                else:
                    if spaceCounter:
                        fenrow.append(str(spaceCounter))
                        spaceCounter = 0
                    fenrow.append(char)
            if spaceCounter:
                fenrow.append(str(spaceCounter))
            fenrows.append("".join(fenrow))

        fen = "/".join(fenrows)
        fen += " "

        # Current color
        fen += fields[8].lower()
        fen += " "

        # Castling
        if fields[10:14] == ["0", "0", "0", "0"]:
            fen += "-"
        else:
            if fields[10] == "1":
                fen += castleSigns[0].upper()
            if fields[11] == "1":
                fen += castleSigns[1].upper()
            if fields[12] == "1":
                fen += castleSigns[0].lower()
            if fields[13] == "1":
                fen += castleSigns[1].lower()
        fen += " "
        # 1 0 1 1 when short castling k1 last possibility

        # En passant
        if fields[9] == "-1":
            fen += "-"
        else:
            fen += fileToEpcord[1 - curcol][int(fields[9])]
        fen += " "

        # Half move clock
        fen += str(max(int(fields[14]), 0))
        fen += " "

        # Standard chess numbering
        fen += fields[25]

        return (
            gameno,
            relation,
            curcol,
            ply,
            wname,
            bname,
            wms,
            bms,
            gain,
            lastmove,
            fen,
        )

    async def onStyle12(self, match):
        style12 = match.groups()[0]
        log.debug("onStyle12: %s" % style12)
        gameno = int(style12.split()[15])
        if gameno in self.queuedStyle12s:
            self.queuedStyle12s[gameno].append(style12)
            return

        try:
            await self.gamemodelStartedEvents[gameno].wait()
        except KeyError:
            pass

        if gameno in self.castleSigns:
            castleSigns = self.castleSigns[gameno]
        else:
            castleSigns = ("k", "q")
        (
            gameno,
            relation,
            curcol,
            ply,
            wname,
            bname,
            wms,
            bms,
            gain,
            lastmove,
            fen,
        ) = self.parseStyle12(style12, castleSigns)

        # examine starts with a <12> line only
        if lastmove is None and relation == IC_POS_EXAMINATING:
            pgnHead = [
                ("Event", "FICS examined game"),
                ("Site", "freechess.org"),
                ("White", wname),
                ("Black", bname),
                ("Result", "*"),
                ("SetUp", "1"),
                ("FEN", fen),
            ]
            pgn = "\n".join(['[%s "%s"]' % line for line in pgnHead]) + "\n*\n"
            wplayer = self.connection.players.get(wname)
            bplayer = self.connection.players.get(bname)

            if self.connection.examined_game is None:
                # examine an archived game from GUI
                if self.connection.archived_examine is not None:
                    no_smoves = False
                    game = self.connection.archived_examine
                    game.gameno = int(gameno)
                    game.relation = relation
                    # game.game_type = GAME_TYPES["examined"]
                    log.debug(
                        "Start examine an existing game by %s" % style12,
                        extra={"task": (self.connection.username, "BM.onStyle12")},
                    )
                else:
                    # examine from console or got mexamine in observed game
                    no_smoves = True
                    game = FICSGame(
                        wplayer,
                        bplayer,
                        gameno=int(gameno),
                        game_type=GAME_TYPES["examined"],
                        minutes=0,
                        inc=0,
                        board=FICSBoard(0, 0, pgn=pgn),
                        relation=relation,
                    )
                    log.debug(
                        "Start new examine game by %s" % style12,
                        extra={"task": (self.connection.username, "BM.onStyle12")},
                    )

                game = self.connection.games.get(game)
                self.connection.examined_game = game

                # don't start another new game when someone (some human examiner or lecturebot/endgamebot)
                # changes our relation in an already started game from IC_POS_OBSERVING_EXAMINATION to
                # IC_POS_EXAMINATING
                if game.relation == IC_POS_OBSERVING_EXAMINATION:
                    game.relation = relation  # IC_POS_EXAMINATING
                    # print("IC_POS_OBSERVING_EXAMINATION --> IC_POS_EXAMINATING")
                    # before this change server sent an unobserve to us
                    # and it removed gameno from started events dict
                    # we have to put it back...
                    self.gamemodelStartedEvents[game.gameno] = asyncio.Event()
                    return

            else:
                # don't start new game in puzzlebot/endgamebot when they just reuse gameno
                log.debug(
                    "emit('boardSetup') with {} {} {} {}".format(
                        gameno, fen, wname, bname
                    ),
                    extra={"task": (self.connection.username, "BM.onStyle12")},
                )
                self.emit("boardSetup", gameno, fen, wname, bname)
                return

            game.relation = relation
            game.board = FICSBoard(0, 0, pgn=pgn)
            self.gamesImObserving[game] = wms, bms

            # start a new game now or after smoves
            self.gamemodelStartedEvents[game.gameno] = asyncio.Event()
            if no_smoves:
                log.debug(
                    "emit('exGameCreated')",
                    extra={"task": (self.connection.username, "BM.onStyle12")},
                )
                self.emit("exGameCreated", game)
                await self.gamemodelStartedEvents[game.gameno].wait()
            else:
                log.debug(
                    "send 'smoves' command",
                    extra={"task": (self.connection.username, "BM.onStyle12")},
                )
                if isinstance(game, FICSHistoryGame):
                    self.connection.client.run_command(
                        "smoves {} {}".format(
                            self.connection.history_owner, game.history_no
                        )
                    )
                elif isinstance(game, FICSJournalGame):
                    self.connection.client.run_command(
                        "smoves {} %{}".format(
                            self.connection.journal_owner, game.journal_no
                        )
                    )
                elif isinstance(game, FICSAdjournedGame):
                    self.connection.client.run_command(
                        "smoves {} {}".format(
                            self.connection.stored_owner, game.opponent.name
                        )
                    )
                self.connection.client.run_command("forward 999")
        else:
            if gameno in self.connection.games.games_by_gameno:
                game = self.connection.games.get_game_by_gameno(gameno)
                if wms < 0 or bms < 0:
                    # fics resend latest style12 line again when one player lost on time
                    return
                if lastmove is None:
                    log.debug(
                        "emit('boardSetup') with {} {} {} {}".format(
                            gameno, fen, wname, bname
                        ),
                        extra={"task": (self.connection.username, "BM.onStyle12")},
                    )
                    self.emit("boardSetup", gameno, fen, wname, bname)
                else:
                    log.debug(
                        "put move %s into game.move_queue" % lastmove,
                        extra={"task": (self.connection.username, "BM.onStyle12")},
                    )
                    game.move_queue.put_nowait(
                        (gameno, ply, curcol, lastmove, fen, wname, bname, wms, bms)
                    )
            else:
                # In some cases (like lost on time) the last move is resent by FICS
                # but game was already removed from self.connection.games
                log.debug(f"Got {style12} but {gameno} not in connection.games")

    async def onExamineGameCreated(self, matchlist):
        style12 = matchlist[-1].groups()[0]
        gameno = int(style12.split()[15])

        castleSigns = self.generateCastleSigns(style12, GAME_TYPES["examined"])
        self.castleSigns[gameno] = castleSigns
        (
            gameno,
            relation,
            curcol,
            ply,
            wname,
            bname,
            wms,
            bms,
            gain,
            lastmove,
            fen,
        ) = self.parseStyle12(style12, castleSigns)

        pgnHead = [
            ("Event", "FICS examined game"),
            ("Site", "freechess.org"),
            ("White", wname),
            ("Black", bname),
            ("Result", "*"),
            ("SetUp", "1"),
            ("FEN", fen),
        ]
        pgn = "\n".join(['[%s "%s"]' % line for line in pgnHead]) + "\n*\n"
        wplayer = self.connection.players.get(wname)
        bplayer = self.connection.players.get(bname)

        game = FICSGame(
            wplayer,
            bplayer,
            gameno=int(gameno),
            game_type=GAME_TYPES["examined"],
            minutes=0,
            inc=0,
            board=FICSBoard(0, 0, pgn=pgn),
            relation=relation,
        )
        log.debug(
            "Starting a game in examine (scratch) mode.",
            extra={"task": (self.connection.username, "BM.onExamineGameCreated")},
        )
        self.connection.examined_game = game
        game = self.connection.games.get(game)

        game.relation = relation
        game.board = FICSBoard(0, 0, pgn=pgn)
        self.gamesImObserving[game] = wms, bms

        # start a new game now
        self.gamemodelStartedEvents[game.gameno] = asyncio.Event()
        log.debug(
            "emit('exGameCreated')",
            extra={"task": (self.connection.username, "BM.onExamineGameCreated")},
        )
        self.emit("exGameCreated", game)
        await self.gamemodelStartedEvents[game.gameno].wait()

    def onGameModelStarted(self, gameno):
        self.gamemodelStartedEvents[gameno].set()

    def onWasPrivate(self, match):
        # When observable games were added to the list later than the latest
        # full send, private information will not be known.
        gameno = int(match.groups()[0])
        try:
            game = self.connection.games.get_game_by_gameno(gameno)
        except KeyError:
            return
        game.private = True

    onWasPrivate.BLKCMD = BLKCMD_OBSERVE

    def tooManySeeks(self, match):
        self.emit("tooManySeeks")

    tooManySeeks.BLKCMD = BLKCMD_SEEK

    def nonoWhileExamine(self, match):
        self.emit("nonoWhileExamine")

    nonoWhileExamine.BLKCMD = BLKCMD_SEEK

    def matchDeclined(self, match):
        (decliner,) = match.groups()
        decliner = self.connection.players.get(decliner)
        self.emit("matchDeclined", decliner)

    @classmethod
    def generateCastleSigns(cls, style12, game_type):
        if game_type.variant_type == FISCHERRANDOMCHESS:
            backrow = style12.split()[0]
            leftside = backrow.find("r")
            rightside = backrow.find("r", leftside + 1)
            return (reprFile[rightside], reprFile[leftside])
        else:
            return ("k", "q")

    def onPlayGameCreated(self, matchlist):
        log.debug(
            "'%s' '%s' '%s'"
            % (matchlist[0].string, matchlist[1].string, matchlist[-1].string),
            extra={"task": (self.connection.username, "BM.onPlayGameCreated")},
        )
        wname, wrating, bname, brating, rated, match_type, minutes, inc = matchlist[
            0
        ].groups()
        item = 2 if self.connection.USCN else 1
        gameno, wname, bname, rated, match_type = matchlist[item].groups()
        gameno = int(gameno)
        wrating = parseRating(wrating)
        brating = parseRating(brating)
        rated = rated == "rated"
        game_type = GAME_TYPES[match_type]

        wplayer = self.connection.players.get(wname)
        bplayer = self.connection.players.get(bname)
        for player, rating in ((wplayer, wrating), (bplayer, brating)):
            if player.ratings[game_type.rating_type] != rating:
                player.ratings[game_type.rating_type] = rating
                player.emit("ratings_changed", game_type.rating_type, player)

        style12 = matchlist[-1].groups()[0]
        castleSigns = self.generateCastleSigns(style12, game_type)
        self.castleSigns[gameno] = castleSigns
        (
            gameno,
            relation,
            curcol,
            ply,
            wname,
            bname,
            wms,
            bms,
            gain,
            lastmove,
            fen,
        ) = self.parseStyle12(style12, castleSigns)

        game = FICSGame(
            wplayer,
            bplayer,
            gameno=gameno,
            rated=rated,
            game_type=game_type,
            minutes=int(minutes),
            inc=int(inc),
            board=FICSBoard(wms, bms, fen=fen),
        )

        game = self.connection.games.get(game)

        for player in (wplayer, bplayer):
            if player.status != IC_STATUS_PLAYING:
                player.status = IC_STATUS_PLAYING
            if player.game != game:
                player.game = game

        self.theGameImPlaying = game
        self.gamemodelStartedEvents[gameno] = asyncio.Event()
        self.connection.client.run_command("follow")
        self.emit("playGameCreated", game)

    def onMatchingSeekOrGetGame(self, matchlist):
        if matchlist[-1].string.startswith("<12>"):
            for line in matchlist[1:-4]:
                if line.startswith("<sr>"):
                    self.connection.glm.on_seek_remove(sr.match(line))
                elif line.startswith("<pr>"):
                    self.connection.om.onOfferRemove(pr.match(line))
            self.onPlayGameCreated(
                (
                    creating0.match(matchlist[-4]),
                    creating1.match(matchlist[-3]),
                    matchlist[-1],
                )
            )
        else:
            self.connection.glm.on_seek_add(matchlist[-1])

    onMatchingSeekOrGetGame.BLKCMD = BLKCMD_SEEK

    def onInterceptedChallenge(self, matchlist):
        self.onMatchingSeekOrGetGame(matchlist)

    onInterceptedChallenge.BLKCMD = BLKCMD_MATCH

    def parseGame(self, matchlist, gameclass, in_progress=False, gameno=None):
        """
        Parses the header and movelist for an observed or stored game from its
        matchlist (an re.match object) into a gameclass (FICSGame or subclass
        of) object.

        in_progress - should be True for an observed game matchlist, and False
        for stored/adjourned games
        """
        # ################   observed game movelist example:
        #        Movelist for game 64:
        #
        #        Ajido (2281) vs. IMgooeyjim (2068) --- Thu Oct 14, 20:36 PDT 2010
        #        Rated standard match, initial time: 15 minutes, increment: 3 seconds.
        #
        #        Move  Ajido                   IMgooeyjim
        #        ----  ---------------------   ---------------------
        #          1.  d4      (0:00.000)      Nf6     (0:00.000)
        #          2.  c4      (0:04.061)      g6      (0:00.969)
        #          3.  Nc3     (0:13.280)      Bg7     (0:06.422)
        #              {Still in progress} *
        #
        # #################   stored game example:
        #        BwanaSlei (1137) vs. mgatto (1336) --- Wed Nov  5, 20:56 PST 2008
        #        Rated blitz match, initial time: 5 minutes, increment: 0 seconds.
        #
        #        Move  BwanaSlei               mgatto
        #        ----  ---------------------   ---------------------
        #        1.  e4      (0:00.000)      c5      (0:00.000)
        #        2.  d4      (0:05.750)      cxd4    (0:03.020)
        #        ...
        #        23.  Qxf3    (1:05.500)
        #             {White lost connection; game adjourned} *
        #
        # ################# stored wild/3 game with style12:
        #        kurushi (1626) vs. mgatto (1627) --- Thu Nov  4, 10:33 PDT 2010
        #        Rated wild/3 match, initial time: 3 minutes, increment: 0 seconds.
        #
        #        <12> nqbrknrn pppppppp -------- -------- -------- -------- PPPPPPPP NQBRKNRN W -1 0 0 0 0 0 17 kurushi mgatto -4 3 0 39 39 169403 45227 1 none (0:00.000) none 0 1 0
        #
        #        Move  kurushi                 mgatto
        #        ----  ---------------------   ---------------------
        #          1.  Nb3     (0:00.000)      d5      (0:00.000)
        #          2.  Nhg3    (0:00.386)      e5      (0:03.672)
        #         ...
        #         28.  Rxd5    (0:00.412)
        #              {Black lost connection; game adjourned} *
        #
        # #################  stored game movelist following stored game(s):
        #        Stored games for mgatto:
        #        C Opponent       On Type          Str  M    ECO Date
        #        1: W BabyLurking     Y [ br  5   0] 29-13 W27  D37 Fri Nov  5, 04:41 PDT 2010
        #        2: W gbtami          N [ wr  5   0] 32-34 W14  --- Thu Oct 21, 00:14 PDT 2010
        #
        #        mgatto (1233) vs. BabyLurking (1455) --- Fri Nov  5, 04:33 PDT 2010
        #        Rated blitz match, initial time: 5 minutes, increment: 0 seconds.
        #
        #        Move  mgatto             BabyLurking
        #        ----  ----------------   ----------------
        #        1.  Nf3     (0:00)     d5      (0:00)
        #        2.  d4      (0:03)     Nf6     (0:00)
        #        3.  c4      (0:03)     e6      (0:00)
        #        {White lost connection; game adjourned} *
        #
        # ################## stored game movelist following stored game(s):
        # ##   Note: A wild stored game in this format won't be parseable into a board because
        # ##   it doesn't come with a style12 that has the start position, so we warn and return
        # ##################
        #        Stored games for mgatto:
        #        C Opponent       On Type          Str  M    ECO Date
        #        1: W gbtami          N [ wr  5   0] 32-34 W14  --- Thu Oct 21, 00:14 PDT 2010
        #
        #        mgatto (1627) vs. gbtami (1881) --- Thu Oct 21, 00:10 PDT 2010
        #        Rated wild/fr match, initial time: 5 minutes, increment: 0 seconds.
        #
        #        Move  mgatto             gbtami
        #        ----  ----------------   ----------------
        #        1.  d4      (0:00)     b6      (0:00)
        #        2.  b3      (0:06)     d5      (0:03)
        #        3.  c4      (0:08)     e6      (0:03)
        #        4.  e3      (0:04)     dxc4    (0:02)
        #        5.  bxc4    (0:02)     g6      (0:09)
        #        6.  Nd3     (0:12)     Bg7     (0:02)
        #        7.  Nc3     (0:10)     Ne7     (0:03)
        #        8.  Be2     (0:08)     c5      (0:05)
        #        9.  a4      (0:07)     cxd4    (0:38)
        #        10.  exd4    (0:06)     Bxd4    (0:03)
        #        11.  O-O     (0:10)     Qc6     (0:06)
        #        12.  Bf3     (0:16)     Qxc4    (0:04)
        #        13.  Bxa8    (0:03)     Rxa8    (0:14)
        #        {White lost connection; game adjourned} *
        #
        # #################   other reasons the game could be stored/adjourned:
        #        Game courtesyadjourned by (Black|White)
        #        Still in progress                    # This one must be a FICS bug
        #        Game adjourned by mutual agreement
        #        (White|Black) lost connection; game adjourned
        #        Game adjourned by ((server shutdown)|(adjudication)|(simul holder))

        index = 0
        if in_progress:
            gameno = int(matchlist[index].groups()[0])
            index += 2
        header1 = (
            matchlist[index]
            if isinstance(matchlist[index], str)
            else matchlist[index].group()
        )

        matches = moveListHeader1.match(header1).groups()
        wname, wrating, bname, brating = matches[:4]
        if self.connection.FatICS:
            year, month, day, hour, minute, timezone = matches[11:]
        else:
            weekday, month, day, hour, minute, timezone, year = matches[4:11]
            month = months.index(month) + 1

        wrating = parseRating(wrating)
        brating = parseRating(brating)
        rated, game_type, minutes, increment = moveListHeader2.match(
            matchlist[index + 1]
        ).groups()
        minutes = int(minutes)
        increment = int(increment)
        game_type = GAME_TYPES[game_type]

        reason = matchlist[-1].group().lower()
        if in_progress:
            result = None
            result_str = "*"
        elif "1-0" in reason:
            result = WHITEWON
            result_str = "1-0"
        elif "0-1" in reason:
            result = BLACKWON
            result_str = "0-1"
        elif "1/2-1/2" in reason:
            result = DRAW
            result_str = "1/2-1/2"
        else:
            result = ADJOURNED
            result_str = "*"
        result, reason = parse_reason(result, reason, wname=wname)

        index += 3
        if matchlist[index].startswith("<12>"):
            style12 = matchlist[index][5:]
            castleSigns = self.generateCastleSigns(style12, game_type)
            (
                gameno,
                relation,
                curcol,
                ply,
                wname,
                bname,
                wms,
                bms,
                gain,
                lastmove,
                fen,
            ) = self.parseStyle12(style12, castleSigns)
            initialfen = fen
            movesstart = index + 4
        else:
            if game_type.rating_type == TYPE_WILD:
                # we need a style12 start position to correctly parse a wild/* board
                log.error(
                    "BoardManager.parseGame: no style12 for %s board."
                    % game_type.fics_name
                )
                return None
            castleSigns = ("k", "q")
            initialfen = None
            movesstart = index + 2

        if in_progress:
            self.castleSigns[gameno] = castleSigns

        moves = {}
        times = {}
        wms = bms = minutes * 60 * 1000

        for line in matchlist[movesstart:-1]:
            if not moveListMoves.match(line):
                log.error('BoardManager.parseGame: unmatched line: "%s"' % repr(line))
                raise Exception(
                    'BoardManager.parseGame: unmatched line: "%s"' % repr(line)
                )
            (
                moveno,
                wmove,
                whour,
                wmin,
                wsec,
                wmsec,
                bmove,
                bhour,
                bmin,
                bsec,
                bmsec,
            ) = moveListMoves.match(line).groups()
            whour = 0 if whour is None else int(whour[0])
            bhour = 0 if bhour is None else int(bhour[0])
            ply = int(moveno) * 2 - 2
            if wmove:
                moves[ply] = wmove
                wms -= (
                    (int(whour) * 60 * 60 * 1000)
                    + (int(wmin) * 60 * 1000)
                    + (int(wsec) * 1000)
                )
                if wmsec is not None:
                    wms -= int(wmsec)
                else:
                    wmsec = 0
                if increment > 0:
                    wms += increment * 1000
                times[ply] = "%01d:%02d:%02d.%03d" % (
                    int(whour),
                    int(wmin),
                    int(wsec),
                    int(wmsec),
                )
            if bmove:
                moves[ply + 1] = bmove
                bms -= (
                    (int(bhour) * 60 * 60 * 1000)
                    + (int(bmin) * 60 * 1000)
                    + (int(bsec) * 1000)
                )
                if bmsec is not None:
                    bms -= int(bmsec)
                else:
                    bmsec = 0
                if increment > 0:
                    bms += increment * 1000
                times[ply + 1] = "%01d:%02d:%02d.%03d" % (
                    int(bhour),
                    int(bmin),
                    int(bsec),
                    int(bmsec),
                )

        if in_progress and gameno in self.queuedStyle12s:
            # Apply queued board updates
            for style12 in self.queuedStyle12s[gameno]:
                (
                    gameno,
                    relation,
                    curcol,
                    ply,
                    wname,
                    bname,
                    wms,
                    bms,
                    gain,
                    lastmove,
                    fen,
                ) = self.parseStyle12(style12, castleSigns)
                if lastmove is None:
                    continue
                moves[ply - 1] = lastmove
                # Updated the queuedMoves in case there has been a takeback
                for moveply in list(moves.keys()):
                    if moveply > ply - 1:
                        del moves[moveply]
            del self.queuedStyle12s[gameno]

        pgnHead = [
            ("Event", f"FICS {rated.lower()} {game_type.fics_name} game"),
            ("Site", "freechess.org"),
            ("White", wname),
            ("Black", bname),
            ("TimeControl", "%d+%d" % (minutes * 60, increment)),
            ("Result", result_str),
            ("WhiteClock", msToClockTimeTag(wms)),
            ("BlackClock", msToClockTimeTag(bms)),
        ]
        if wrating != 0:
            pgnHead += [("WhiteElo", wrating)]
        if brating != 0:
            pgnHead += [("BlackElo", brating)]
        if year and month and day and hour and minute:
            pgnHead += [
                ("Date", "%04d.%02d.%02d" % (int(year), int(month), int(day))),
                ("Time", "%02d:%02d:00" % (int(hour), int(minute))),
            ]
        if initialfen:
            pgnHead += [("SetUp", "1"), ("FEN", initialfen)]
        if game_type.variant_type == FISCHERRANDOMCHESS:
            pgnHead += [("Variant", "Fischerandom")]
            # FR is the only variant used in this tag by the PGN generator @
            # ficsgames.org. They put all the other wild/* stuff only in the
            # "Event" header.
        elif game_type.variant_type == CRAZYHOUSECHESS:
            pgnHead += [("Variant", "Crazyhouse")]
        elif game_type.variant_type in (WILDCASTLECHESS, WILDCASTLESHUFFLECHESS):
            pgnHead += [("Variant", "Wildcastle")]
        elif game_type.variant_type == ATOMICCHESS:
            pgnHead += [("Variant", "Atomic")]
        elif game_type.variant_type == LOSERSCHESS:
            pgnHead += [("Variant", "Losers")]
        elif game_type.variant_type == SUICIDECHESS:
            pgnHead += [("Variant", "Suicide")]
        elif game_type.variant_type == GIVEAWAYCHESS:
            pgnHead += [("Variant", "Giveaway")]
        pgn = "\n".join(['[%s "%s"]' % line for line in pgnHead]) + "\n"

        moves = sorted(moves.items())
        for ply, move in moves:
            if ply % 2 == 0:
                pgn += "%d. " % (ply // 2 + 1)
            time = times[ply]
            pgn += f"{move} {{[%emt {time}]}} "
        pgn += "*\n"

        wplayer = self.connection.players.get(wname)
        bplayer = self.connection.players.get(bname)
        for player, rating in ((wplayer, wrating), (bplayer, brating)):
            if player.ratings[game_type.rating_type] != rating:
                player.ratings[game_type.rating_type] = rating
                player.emit("ratings_changed", game_type.rating_type, player)
        game = gameclass(
            wplayer,
            bplayer,
            game_type=game_type,
            result=result,
            rated=(rated.lower() == "rated"),
            minutes=minutes,
            inc=increment,
            board=FICSBoard(wms, bms, pgn=pgn),
        )

        if in_progress:
            game.gameno = gameno
        else:
            if gameno is not None:
                game.gameno = gameno
            game.reason = reason
        game = self.connection.games.get(game, emit=False)

        return game

    def on_game_remove(self, match):
        gameno, wname, bname, comment, result = match.groups()
        result, reason = parse_reason(reprResult.index(result), comment, wname=wname)

        try:
            wplayer = self.connection.players.get(wname)
            wplayer.restore_previous_status()
            # no status update will be sent by
            # FICS if the player doesn't become available, so we restore
            # previous status first (not necessarily true, but the best guess)
        except KeyError:
            print("%s not in self.connections.players - creating" % wname)
            wplayer = FICSPlayer(wname)

        try:
            bplayer = self.connection.players.get(bname)
            bplayer.restore_previous_status()
        except KeyError:
            print("%s not in self.connections.players - creating" % bname)
            bplayer = FICSPlayer(bname)

        game = FICSGame(
            wplayer, bplayer, gameno=int(gameno), result=result, reason=reason
        )
        if wplayer.game is not None:
            game.rated = wplayer.game.rated
        game = self.connection.games.get(game, emit=False)
        self.connection.games.game_ended(game)
        # Do this last to give anybody connected to the game's signals a chance
        # to disconnect from them first
        wplayer.game = None
        bplayer.game = None

    async def onObserveGameCreated(self, matchlist):
        log.debug(
            "'%s'" % (matchlist[1].string),
            extra={"task": (self.connection.username, "BM.onObserveGameCreated")},
        )
        if self.connection.USCN:
            # TODO? is this ok?
            game_type = GAME_TYPES["blitz"]
            castleSigns = ("k", "q")
        else:
            (
                gameno,
                wname,
                wrating,
                bname,
                brating,
                rated,
                gametype,
                minutes,
                inc,
            ) = matchlist[1].groups()
            wrating = parseRating(wrating)
            brating = parseRating(brating)
            game_type = GAME_TYPES[gametype]

        style12 = matchlist[-1].groups()[0]

        castleSigns = self.generateCastleSigns(style12, game_type)
        (
            gameno,
            relation,
            curcol,
            ply,
            wname,
            bname,
            wms,
            bms,
            gain,
            lastmove,
            fen,
        ) = self.parseStyle12(style12, castleSigns)
        gameno = int(gameno)
        self.castleSigns[gameno] = castleSigns

        wplayer = self.connection.players.get(wname)
        bplayer = self.connection.players.get(bname)

        if relation == IC_POS_OBSERVING_EXAMINATION:
            pgnHead = [
                ("Event", f"FICS {rated} {game_type.fics_name} game"),
                ("Site", "freechess.org"),
                ("White", wname),
                ("Black", bname),
                ("Result", "*"),
                ("SetUp", "1"),
                ("FEN", fen),
            ]
            pgn = "\n".join(['[%s "%s"]' % line for line in pgnHead]) + "\n*\n"
            game = FICSGame(
                wplayer,
                bplayer,
                gameno=gameno,
                rated=rated == "rated",
                game_type=game_type,
                minutes=int(minutes),
                inc=int(inc),
                board=FICSBoard(wms, bms, pgn=pgn),
                relation=relation,
            )
            game = self.connection.games.get(game)

            # when puzzlebot reuses same gameno for starting next puzzle
            # sometimes no unexamine sent by server, so we have to set None to
            # self.connection.examined_game to guide self.onStyle12() a bit...
            if (
                self.connection.examined_game is not None
                and self.connection.examined_game.gameno == gameno
            ):
                log.debug(
                    "BM.onObserveGameCreated: exGameReset emitted; self.connection.examined_game = %s"
                    % gameno
                )
                self.emit("exGameReset", self.connection.examined_game)
                self.connection.examined_game = None

            game.relation = relation  # IC_POS_OBSERVING_EXAMINATION
            self.gamesImObserving[game] = wms, bms

            self.gamemodelStartedEvents[game.gameno] = asyncio.Event()
            # puzzlebot sometimes creates next puzzle with same wplayer,bplayer,gameno
            game.move_queue = asyncio.Queue()
            self.emit("obsGameCreated", game)
            await self.gamemodelStartedEvents[game.gameno].wait()
        else:
            game = FICSGame(
                wplayer,
                bplayer,
                gameno=gameno,
                rated=rated == "rated",
                game_type=game_type,
                minutes=int(minutes),
                inc=int(inc),
                relation=relation,
            )
            game = self.connection.games.get(game, emit=False)

            if not game.supported:
                log.warning(
                    "Trying to follow an unsupported type game %s" % game.game_type
                )
                return

            if game.gameno in self.gamemodelStartedEvents:
                log.warning("%s already in gamemodelstartedevents" % game.gameno)
                return

            self.gamesImObserving[game] = wms, bms
            self.queuedStyle12s[game.gameno] = []
            self.queuedEmits[game.gameno] = []
            self.gamemodelStartedEvents[game.gameno] = asyncio.Event()

            # FICS doesn't send the move list after 'observe' and 'follow' commands
            self.connection.client.run_command("moves %d" % game.gameno)

    onObserveGameCreated.BLKCMD = BLKCMD_OBSERVE

    async def onObserveGameMovesReceived(self, matchlist):
        log.debug(
            "'%s'" % (matchlist[0].string),
            extra={"task": (self.connection.username, "BM.onObserveGameMovesReceived")},
        )
        game = self.parseGame(matchlist, FICSGame, in_progress=True)
        if game.gameno not in self.gamemodelStartedEvents:
            return
        if game.gameno not in self.queuedEmits:
            return
        self.emit("obsGameCreated", game)
        try:
            await self.gamemodelStartedEvents[game.gameno].wait()
        except KeyError:
            pass

        for emit in self.queuedEmits[game.gameno]:
            emit()
        del self.queuedEmits[game.gameno]

        wms, bms = self.gamesImObserving[game]
        self.emit("timesUpdate", game.gameno, wms, bms)

    onObserveGameMovesReceived.BLKCMD = BLKCMD_MOVES

    async def onArchiveGameSMovesReceived(self, matchlist):
        log.debug(
            "'%s'" % (matchlist[0].string),
            extra={
                "task": (self.connection.username, "BM.onArchiveGameSMovesReceived")
            },
        )
        klass = (
            FICSAdjournedGame if "adjourn" in matchlist[-1].group() else FICSHistoryGame
        )
        if self.connection.examined_game is not None:
            gameno = self.connection.examined_game.gameno
        else:
            gameno = None
        game = self.parseGame(matchlist, klass, in_progress=False, gameno=gameno)
        if game.gameno not in self.gamemodelStartedEvents:
            self.emit("archiveGamePreview", game)
            return
        game.relation = IC_POS_EXAMINATING
        game.game_type = GAME_TYPES["examined"]
        self.emit("exGameCreated", game)
        try:
            await self.gamemodelStartedEvents[game.gameno].wait()
        except KeyError:
            pass

    onArchiveGameSMovesReceived.BLKCMD = BLKCMD_SMOVES

    def onGameEnd(self, games, game):
        log.debug("BM.onGameEnd: %s" % game)
        if game == self.theGameImPlaying:
            # if game.gameno in self.gamemodelStartedEvents:
            # await self.gamemodelStartedEvents[game.gameno].wait()
            self.emit("curGameEnded", game)
            self.theGameImPlaying = None
            if game.gameno in self.gamemodelStartedEvents:
                del self.gamemodelStartedEvents[game.gameno]

        elif game in self.gamesImObserving:
            log.debug("BM.onGameEnd: %s: gamesImObserving" % game)
            if game.gameno in self.queuedEmits:
                log.debug("BM.onGameEnd: %s: queuedEmits" % game)
                self.queuedEmits[game.gameno].append(
                    lambda: self.emit("obsGameEnded", game)
                )
            else:
                # if game.gameno in self.gamemodelStartedEvents:
                # await self.gamemodelStartedEvents[game.gameno].wait()
                del self.gamesImObserving[game]
                self.emit("obsGameEnded", game)

    def onGamePause(self, match):
        gameno, state = match.groups()
        gameno = int(gameno)
        if gameno in self.queuedEmits:
            self.queuedEmits[gameno].append(
                lambda: self.emit("gamePaused", gameno, state == "paused")
            )
        else:
            # if gameno in self.gamemodelStartedEvents:
            # await self.gamemodelStartedEvents[gameno].wait()
            self.emit("gamePaused", gameno, state == "paused")

    def onUnobserveGame(self, match):
        gameno = int(match.groups()[0])
        log.debug("BM.onUnobserveGame: gameno: %s" % gameno)
        try:
            del self.gamemodelStartedEvents[gameno]
            game = self.connection.games.get_game_by_gameno(gameno)
        except KeyError:
            return
        self.emit("obsGameUnobserved", game)
        # TODO: delete self.castleSigns[gameno] ?

    onUnobserveGame.BLKCMD = BLKCMD_UNOBSERVE

    def player_lagged(self, match):
        gameno, player, num_seconds = match.groups()
        player = self.connection.players.get(player)
        self.emit("player_lagged", player)

    def opp_not_out_of_time(self, match):
        self.emit("opp_not_out_of_time")

    opp_not_out_of_time.BLKCMD = BLKCMD_FLAG

    def req_not_fit_formula(self, matchlist):
        player, formula = matchlist[1].groups()
        player = self.connection.players.get(player)
        self.emit("req_not_fit_formula", player, formula)

    req_not_fit_formula.BLKCMD = BLKCMD_MATCH

    def player_on_censor(self, match):
        (player,) = match.groups()
        player = self.connection.players.get(player)
        self.emit("player_on_censor", player)

    player_on_censor.BLKCMD = BLKCMD_MATCH

    def player_on_noplay(self, match):
        (player,) = match.groups()
        player = self.connection.players.get(player)
        self.emit("player_on_noplay", player)

    player_on_noplay.BLKCMD = BLKCMD_MATCH

    def made_examined(self, match):
        """Changing from observer to examiner"""
        player, gameno = match.groups()
        gameno = int(gameno)
        try:
            self.connection.games.get_game_by_gameno(gameno)
        except KeyError:
            return
        self.emit("madeExamined", gameno)

    def made_unexamined(self, match):
        """You are no longer examine game"""
        log.debug("BM.made_unexamined(): exGameReset emitted")
        self.emit("exGameReset", self.connection.examined_game)
        self.connection.examined_game = None
        (gameno,) = match.groups()
        gameno = int(gameno)
        try:
            self.connection.games.get_game_by_gameno(gameno)
        except KeyError:
            return
        self.emit("madeUnExamined", gameno)

    ############################################################################
    #   Interacting                                                            #
    ############################################################################

    def isPlaying(self):
        return self.theGameImPlaying is not None

    def sendMove(self, move):
        self.connection.client.run_command(move)

    def resign(self):
        self.connection.client.run_command("resign")

    def callflag(self):
        self.connection.client.run_command("flag")

    def observe(self, game, player=None):
        if game is not None:
            self.connection.client.run_command("observe %d" % game.gameno)
        elif player is not None:
            self.connection.client.run_command("observe %s" % player.name)

    def follow(self, player):
        self.connection.client.run_command("follow %s" % player.name)

    def unexamine(self):
        self.connection.client.run_command("unexamine")

    def unobserve(self, game):
        if game.gameno is not None:
            self.connection.client.run_command("unobserve %d" % game.gameno)

    def play(self, seekno):
        self.connection.client.run_command("play %s" % seekno)

    def accept(self, offerno):
        self.connection.client.run_command("accept %s" % offerno)

    def decline(self, offerno):
        self.connection.client.run_command("decline %s" % offerno)


if __name__ == "__main__":
    from pychess.ic.FICSConnection import Connection

    con = Connection("", "", True, "", "")
    bm = BoardManager(con)

    print(
        bm._BoardManager__parseStyle12(
            "rkbrnqnb pppppppp -------- -------- -------- -------- PPPPPPPP RKBRNQNB W -1 1 1 1 1 0 161 GuestNPFS GuestMZZK -1 2 12 39 39 120 120 1 none (0:00) none 1 0 0",
            ("d", "a"),
        )
    )

    print(
        bm._BoardManager__parseStyle12(
            "rnbqkbnr pppp-ppp -------- ----p--- ----PP-- -------- PPPP--PP RNBQKBNR B 5 1 1 1 1 0 241 GuestGFFC GuestNXMP -4 2 12 39 39 120000 120000 1 none (0:00.000) none 0 0 0",
            ("k", "q"),
        )
    )



import re

from gi.repository import GObject

from pychess.Utils.const import (
    DRAW_OFFER,
    ABORT_OFFER,
    ADJOURN_OFFER,
    TAKEBACK_OFFER,
    PAUSE_OFFER,
    RESUME_OFFER,
    SWITCH_OFFER,
    RESIGNATION,
    FLAG_CALL,
    MATCH_OFFER,
    WHITE,
    ACTION_ERROR_SWITCH_UNDERWAY,
    ACTION_ERROR_CLOCK_NOT_STARTED,
    ACTION_ERROR_CLOCK_NOT_PAUSED,
    ACTION_ERROR_NONE_TO_ACCEPT,
    ACTION_ERROR_NONE_TO_WITHDRAW,
    ACTION_ERROR_NONE_TO_DECLINE,
    ACTION_ERROR_TOO_LARGE_UNDO,
    ACTION_ERROR_NOT_OUT_OF_TIME,
)

from pychess.Utils.Offer import Offer
from pychess.System.Log import log
from pychess.ic import GAME_TYPES, VariantGameType
from pychess.ic.FICSObjects import FICSChallenge

names = r"\w+(?:\([A-Z\*]+\))*"

rated = "(rated|unrated)"
colors = r"(?:\[(white|black)\])?"
ratings = r"\(([0-9\ \-\+]{1,4}[E P]?)\)"
loaded_from = r"(?: Loaded from (wild[/\w]*))?"
adjourned = r"(?: (\(adjourned\)))?"

matchreUntimed = re.compile(
    rf"(\w+) {ratings} {colors} ?(\w+) {ratings} {rated} (untimed)\s*"
)
matchre = re.compile(
    r"(\w+) %s %s ?(\w+) %s %s (\w+) (\d+) (\d+)%s%s"
    % (ratings, colors, ratings, rated, loaded_from, adjourned)
)

# <pf> 39 w=GuestDVXV t=match p=GuestDVXV (----) [black] GuestNXMP (----) unrated blitz 2 12
# <pf> 16 w=GuestDVXV t=match p=GuestDVXV (----) GuestNXMP (----) unrated wild 2 12 Loaded from wild/fr
# <pf> 39 w=GuestDVXV t=match p=GuestDVXV (----) GuestNXMP (----) unrated blitz 2 12 (adjourned)
# <pf> 45 w=GuestGYXR t=match p=GuestGYXR (----) Lobais (----) unrated losers 2 12
# <pf> 45 w=GuestYDDR t=match p=GuestYDDR (----) mgatto (1358) unrated untimed
# <pf> 71 w=joseph t=match p=joseph (1632) mgatto (1742) rated wild 5 1 Loaded from wild/fr (adjourned)
# <pf> 59 w=antiseptic t=match p=antiseptic (1945) mgatto (1729) rated wild 6 1 Loaded from wild/4 (adjourned)
#
# Known offers: abort accept adjourn draw match pause unpause switch takeback
#

strToOfferType = {
    "draw": DRAW_OFFER,
    "abort": ABORT_OFFER,
    "adjourn": ADJOURN_OFFER,
    "takeback": TAKEBACK_OFFER,
    "pause": PAUSE_OFFER,
    "unpause": RESUME_OFFER,
    "switch": SWITCH_OFFER,
    "resign": RESIGNATION,
    "flag": FLAG_CALL,
    "match": MATCH_OFFER,
}

offerTypeToStr = {}
for k, v in strToOfferType.items():
    offerTypeToStr[v] = k


class OfferManager(GObject.GObject):
    __gsignals__ = {
        "onOfferAdd": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "onOfferRemove": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "onOfferDeclined": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "onChallengeAdd": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "onChallengeRemove": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "onActionError": (GObject.SignalFlags.RUN_FIRST, None, (object, int)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)

        self.connection = connection

        self.connection.expect_line(
            self.onOfferAdd, r"<p(t|f)> (\d+) w=%s t=(\w+) p=(.+)" % names
        )
        self.connection.expect_line(self.onOfferRemove, r"<pr> (\d+)")

        for ficsstring, offer, error in (
            (
                "You cannot switch sides once a game is underway.",
                Offer(SWITCH_OFFER),
                ACTION_ERROR_SWITCH_UNDERWAY,
            ),
            (
                "Opponent is not out of time.",
                Offer(FLAG_CALL),
                ACTION_ERROR_NOT_OUT_OF_TIME,
            ),
            (
                "The clock is not ticking yet.",
                Offer(PAUSE_OFFER),
                ACTION_ERROR_CLOCK_NOT_STARTED,
            ),
            (
                "The clock is not ticking.",
                Offer(FLAG_CALL),
                ACTION_ERROR_CLOCK_NOT_STARTED,
            ),
            (
                "The clock is not paused.",
                Offer(RESUME_OFFER),
                ACTION_ERROR_CLOCK_NOT_PAUSED,
            ),
        ):
            self.connection.expect_line(
                lambda match: self.emit("onActionError", offer, error), ficsstring
            )

        self.connection.expect_line(
            self.notEnoughMovesToUndo,
            r"There are (?:(no)|only (\d+) half) moves in your game\.",
        )

        self.connection.expect_line(
            self.noOffersToAccept, "There are no ([^ ]+) offers to (accept)."
        )

        self.connection.expect_line(
            self.onOfferDeclined,
            r"\w+ declines the (draw|takeback|pause|unpause|abort|adjourn) request\.",
        )

        self.lastPly = 0
        self.offers = {}

        self.connection.client.run_command("iset pendinfo 1")

    def onOfferDeclined(self, match):
        log.debug("OfferManager.onOfferDeclined: match.string=%s" % match.string)
        type = match.groups()[0]
        offer = Offer(strToOfferType[type])
        self.emit("onOfferDeclined", offer)

    def noOffersToAccept(self, match):
        offertype, request = match.groups()
        if request == "accept":
            error = ACTION_ERROR_NONE_TO_ACCEPT
        elif request == "withdraw":
            error = ACTION_ERROR_NONE_TO_WITHDRAW
        elif request == "decline":
            error = ACTION_ERROR_NONE_TO_DECLINE
        offer = Offer(strToOfferType[offertype])
        self.emit("onActionError", offer, error)

    def notEnoughMovesToUndo(self, match):
        ply = match.groups()[0] or match.groups()[1]
        if ply == "no":
            ply = 0
        else:
            ply = int(ply)
        offer = Offer(TAKEBACK_OFFER, param=ply)
        self.emit("onActionError", offer, ACTION_ERROR_TOO_LARGE_UNDO)

    def onOfferAdd(self, match):
        log.debug("OfferManager.onOfferAdd: match.string=%s" % match.string)

        tofrom, index, offertype, parameters = match.groups()
        index = int(index)

        if tofrom == "t":
            # ICGameModel keeps track of the offers we've sent ourselves, so we
            # don't need this
            return
        if offertype not in strToOfferType:
            log.warning(
                "OfferManager.onOfferAdd: Declining unknown offer type: "
                + "offertype=%s parameters=%s index=%d" % (offertype, parameters, index)
            )
            self.connection.client.run_command("decline %d" % index)
            return
        offertype = strToOfferType[offertype]
        if offertype == TAKEBACK_OFFER:
            offer = Offer(offertype, param=int(parameters), index=index)
        else:
            offer = Offer(offertype, index=index)
        self.offers[offer.index] = offer

        if offer.type == MATCH_OFFER:
            is_adjourned = False
            if matchreUntimed.match(parameters) is not None:
                fname, frating, col, tname, trating, rated, type = matchreUntimed.match(
                    parameters
                ).groups()
                mins = 0
                incr = 0
                gametype = GAME_TYPES["untimed"]
            else:
                (
                    fname,
                    frating,
                    col,
                    tname,
                    trating,
                    rated,
                    gametype,
                    mins,
                    incr,
                    wildtype,
                    adjourned,
                ) = matchre.match(parameters).groups()
                if (wildtype and "adjourned" in wildtype) or (
                    adjourned and "adjourned" in adjourned
                ):
                    is_adjourned = True
                if wildtype and "wild" in wildtype:
                    gametype = wildtype

                try:
                    gametype = GAME_TYPES[gametype]
                except KeyError:
                    log.warning(
                        "OfferManager.onOfferAdd: auto-declining "
                        + "unknown offer type: '%s'\n" % gametype
                    )
                    self.decline(offer)
                    del self.offers[offer.index]
                    return

            player = self.connection.players.get(fname)
            rating = frating.strip()
            rating = int(rating) if rating.isdigit() else 0
            if player.ratings[gametype.rating_type] != rating:
                player.ratings[gametype.rating_type] = rating
                player.emit("ratings_changed", gametype.rating_type, player)
            rated = rated != "unrated"
            challenge = FICSChallenge(
                index,
                player,
                int(mins),
                int(incr),
                rated,
                col,
                gametype,
                adjourned=is_adjourned,
            )
            self.emit("onChallengeAdd", challenge)

        else:
            log.debug("OfferManager.onOfferAdd: emitting onOfferAdd: %s" % offer)
            self.emit("onOfferAdd", offer)

    def onOfferRemove(self, match):
        log.debug("OfferManager.onOfferRemove: match.string=%s" % match.string)
        index = int(match.groups()[0])
        if index not in self.offers:
            return
        if self.offers[index].type == MATCH_OFFER:
            self.emit("onChallengeRemove", index)
        else:
            self.emit("onOfferRemove", self.offers[index])
        del self.offers[index]

    ###

    def challenge(self, player_name, game_type, startmin, incsec, rated, color=None):
        log.debug(
            "OfferManager.challenge: %s %s %s %s %s %s"
            % (player_name, game_type, startmin, incsec, rated, color)
        )
        rchar = rated and "r" or "u"
        if color is not None:
            cchar = color == WHITE and "w" or "b"
        else:
            cchar = ""
        s = "match %s %d %d %s %s" % (player_name, startmin, incsec, rchar, cchar)
        if isinstance(game_type, VariantGameType):
            s += " " + game_type.seek_text
        self.connection.client.run_command(s)

    def offer(self, offer):
        log.debug("OfferManager.offer: %s" % offer)
        s = offerTypeToStr[offer.type]
        if offer.type == TAKEBACK_OFFER:
            s += " " + str(offer.param)
        self.connection.client.run_command(s)

    ###

    def withdraw(self, offer):
        log.debug("OfferManager.withdraw: %s" % offer)
        self.connection.client.run_command("withdraw t %s" % offerTypeToStr[offer.type])

    def accept(self, offer):
        log.debug("OfferManager.accept: %s" % offer)
        if offer.index is not None:
            self.acceptIndex(offer.index)
        else:
            self.connection.client.run_command(
                "accept t %s" % offerTypeToStr[offer.type]
            )

    def decline(self, offer):
        log.debug("OfferManager.decline: %s" % offer)
        if offer.index is not None:
            self.declineIndex(offer.index)
        else:
            self.connection.client.run_command(
                "decline t %s" % offerTypeToStr[offer.type]
            )

    def acceptIndex(self, index):
        log.debug("OfferManager.acceptIndex: index=%s" % index)
        self.connection.client.run_command("accept %s" % index)

    def declineIndex(self, index):
        log.debug("OfferManager.declineIndex: index=%s" % index)
        self.connection.client.run_command("decline %s" % index)

    def playIndex(self, index):
        log.debug("OfferManager.playIndex: index=%s" % index)
        self.connection.client.run_command("play %s" % index)



from urllib.request import urlopen
from urllib.parse import unquote

from gi.repository import Gtk

from pychess.perspectives import perspective_manager


class RecentChooserMenu(Gtk.RecentChooserMenu):
    def __init__(self):
        Gtk.RecentChooserMenu.__init__(self)

        def recent_item_activated(self):
            uri = self.get_current_uri()
            try:
                urlopen(unquote(uri)).close()
                perspective = perspective_manager.get_perspective("database")
                perspective.open_chessfile(self.get_current_uri())
            except OSError:
                # shomething wrong whit the uri
                recent_manager.remove_item(uri)

        self.set_show_tips(True)
        self.set_sort_type(Gtk.RecentSortType.MRU)
        self.set_limit(10)
        self.set_name("recent_menu")

        file_filter = Gtk.RecentFilter()
        file_filter.add_mime_type("application/x-chess-pgn")
        file_filter.add_mime_type("application/x-chess-epd")
        file_filter.add_mime_type("application/x-chess-fen")
        file_filter.add_pattern("*.pgn")
        file_filter.add_pattern("*.epd")
        file_filter.add_pattern("*.fen")
        self.set_filter(file_filter)

        self.connect("item-activated", recent_item_activated)


recent_manager = Gtk.RecentManager.get_default()
recent_menu = RecentChooserMenu()



import platform
import sys


def get_cpu():
    result = {}

    # Operating system
    result["platform"] = sys.platform
    result["windows"] = result["platform"] == "win32"
    result["linux"] = result["platform"].startswith("linux")
    result["mac"] = result["platform"] == "darwin"
    result["release"] = platform.release().lower()
    result["binext"] = ".exe" if result["windows"] else ""

    # Number of bits
    result["bitness"] = (
        "64" if platform.machine().endswith("64") or sys.maxsize > 2**32 else "32"
    )

    # Instruction sets
    try:
        info = ""
        with open("/proc/cpuinfo") as f:
            info = f.read()
        result["popcnt"] = "popcnt" in info
        result["bmi2"] = "bmi2" in info
    except OSError:
        # Logic not fully true
        guess = result["bitness"] == "64"
        result["popcnt"] = guess
        result["bmi2"] = guess
    return result



import colorsys
import sys
import xml.etree.ElementTree as ET

# from io import BytesIO

import gi

try:
    gi.require_version("Gdk", "3.0")
    gi.require_version("GObject", "2.0")
    gi.require_version("Gtk", "3.0")
    gi.require_version("Pango", "1.0")
    from gi.repository import Gtk, Gdk, GObject, Pango

    gi.require_version("GdkPixbuf", "2.0")
    from gi.repository.GdkPixbuf import Pixbuf
except Exception:
    print("Failed to import required gi module version")
    sys.exit(1)

from pychess.System import conf
from pychess.System.Log import log
from pychess.System.prefix import addDataPrefix


def createCombo(combo, data=[], name=None, ellipsize_mode=None):
    if name is not None:
        combo.set_name(name)
    lst_store = Gtk.ListStore(Pixbuf, str)
    for row in data:
        lst_store.append(row)
    combo.clear()

    combo.set_model(lst_store)
    crp = Gtk.CellRendererPixbuf()
    crp.set_property("xalign", 0)
    crp.set_property("xpad", 2)
    combo.pack_start(crp, False)
    combo.add_attribute(crp, "pixbuf", 0)

    crt = Gtk.CellRendererText()
    crt.set_property("xalign", 0)
    crt.set_property("xpad", 4)
    combo.pack_start(crt, True)
    combo.add_attribute(crt, "text", 1)
    if ellipsize_mode is not None:
        crt.set_property("ellipsize", ellipsize_mode)


def updateCombo(combo, data):
    def get_active(combobox):
        model = combobox.get_model()
        active = combobox.get_active()
        if active < 0:
            return None
        return model[active][1]

    last_active = get_active(combo)
    lst_store = combo.get_model()
    lst_store.clear()
    new_active = 0
    for i, row in enumerate(data):
        lst_store.append(row)
        if last_active == row[1]:
            new_active = i
    combo.set_active(new_active)


def genColor(n, startpoint=0):
    assert n >= 1
    # This splits the 0 - 1 segment in the pizza way
    hue = (2 * n - 1) / (2.0 ** (n - 1).bit_length()) - 1
    hue = (hue + startpoint) % 1
    # We set saturation based on the amount of green, scaled to the interval
    # [0.6..0.8]. This ensures a consistent lightness over all colors.
    rgb = colorsys.hsv_to_rgb(hue, 1, 1)
    rgb = colorsys.hsv_to_rgb(hue, 1, (1 - rgb[1]) * 0.2 + 0.6)
    # This algorithm ought to balance colors more precisely, but it overrates
    # the lightness of yellow, and nearly makes it black
    # yiq = colorsys.rgb_to_yiq(*rgb)
    # rgb = colorsys.yiq_to_rgb(.125, yiq[1], yiq[2])
    return rgb


def keepDown(scrolledWindow):
    def changed(vadjust):
        if not hasattr(vadjust, "need_scroll") or vadjust.need_scroll:
            vadjust.set_value(vadjust.get_upper() - vadjust.get_page_size())
            vadjust.need_scroll = True

    scrolledWindow.get_vadjustment().connect("changed", changed)

    def value_changed(vadjust):
        vadjust.need_scroll = (
            abs(vadjust.get_value() + vadjust.get_page_size() - vadjust.get_upper())
            < vadjust.get_step_increment()
        )

    scrolledWindow.get_vadjustment().connect("value-changed", value_changed)


# wrap analysis text column. thanks to
# http://www.islascruz.org/html/index.php?blog/show/Wrap-text-in-a-TreeView-column.html
def appendAutowrapColumn(treeview, name, **kvargs):
    cell = Gtk.CellRendererText()
    # cell.props.wrap_mode = Pango.WrapMode.WORD
    # TODO:
    # changed to ellipsize instead until "never ending grow" bug gets fixed
    # see https://github.com/pychess/pychess/issues/1054
    cell.props.ellipsize = Pango.EllipsizeMode.END
    column = Gtk.TreeViewColumn(name, cell, **kvargs)
    treeview.append_column(column)

    def callback(treeview, allocation, column, cell):
        otherColumns = [c for c in treeview.get_columns() if c != column]
        newWidth = allocation.width - sum(c.get_width() for c in otherColumns)

        hsep = GObject.Value()
        hsep.init(GObject.TYPE_INT)
        hsep.set_int(0)
        treeview.style_get_property("horizontal-separator", hsep)
        newWidth -= hsep.get_int() * (len(otherColumns) + 1) * 2
        if cell.props.wrap_width == newWidth or newWidth <= 0:
            return
        cell.props.wrap_width = newWidth
        store = treeview.get_model()
        store_iter = store.get_iter_first()
        while store_iter and store.iter_is_valid(store_iter):
            store.row_changed(store.get_path(store_iter), store_iter)
            store_iter = store.iter_next(store_iter)
        treeview.set_size_request(0, -1)

    # treeview.connect_after("size-allocate", callback, column, cell)

    scroll = treeview.get_parent()
    if isinstance(scroll, Gtk.ScrolledWindow):
        scroll.set_policy(Gtk.PolicyType.NEVER, scroll.get_policy()[1])

    return cell


METHODS = (
    # Gtk.SpinButton should be listed prior to Gtk.Entry, as it is a
    # subclass, but requires different handling
    (Gtk.SpinButton, ("get_value", "set_value", "value-changed")),
    (Gtk.Entry, ("get_text", "set_text", "changed")),
    (Gtk.Expander, ("get_expanded", "set_expanded", "notify::expanded")),
    (Gtk.ComboBox, ("get_active", "set_active", "changed")),
    (Gtk.IconView, ("_get_active", "_set_active", "selection-changed")),
    (Gtk.ToggleButton, ("get_active", "set_active", "toggled")),
    (Gtk.CheckMenuItem, ("get_active", "set_active", "toggled")),
    (Gtk.Range, ("get_value", "set_value", "value-changed")),
    (Gtk.TreeSortable, ("get_value", "set_value", "sort-column-changed")),
    (Gtk.Paned, ("get_position", "set_position", "notify::position")),
)


def keep(widget, key, get_value_=None, set_value_=None):  # , first_value=None):
    if widget is None:
        raise AttributeError("key '%s' isn't in widgets" % key)

    for class_, methods_ in METHODS:
        # Use try-except just to make spinx happy...
        try:
            if isinstance(widget, class_):
                getter, setter, signal = methods_
                break
        except TypeError:
            getter, setter, signal = methods_
            break
    else:
        raise AttributeError("I don't have any knowledge of type: '%s'" % widget)

    if get_value_:

        def get_value():
            return get_value_(widget)

    else:
        get_value = getattr(widget, getter)

    if set_value_:

        def set_value(v):
            return set_value_(widget, v)

    else:
        set_value = getattr(widget, setter)

    def setFromConf():
        try:
            v = conf.get(key)
        except TypeError:
            log.warning(
                "uistuff.keep.setFromConf: Key '%s' from conf had the wrong type '%s', ignored"
                % (key, type(conf.get(key)))
            )
            # print("uistuff.keep TypeError %s %s" % (key, conf.get(key)))
        else:
            set_value(v)

    def callback(*args):
        if not conf.hasKey(key) or conf.get(key) != get_value():
            conf.set(key, get_value())

    widget.connect(signal, callback)
    conf.notify_add(key, lambda *args: setFromConf())

    if conf.hasKey(key):
        setFromConf()
    elif conf.get(key) is not None:
        conf.set(key, conf.get(key))


# loadDialogWidget() and saveDialogWidget() are similar to uistuff.keep() but are needed
# for saving widget values for Gtk.Dialog instances that are loaded with different
# sets of values/configurations and which also aren't instant save like in
# uistuff.keep(), but rather are saved later if and when the user clicks
# the dialog's OK button
def loadDialogWidget(
    widget,
    widget_name,
    config_number,
    get_value_=None,
    set_value_=None,
    first_value=None,
):
    key = widget_name + "-" + str(config_number)

    if widget is None:
        raise AttributeError("key '%s' isn't in widgets" % widget_name)

    for class_, methods_ in METHODS:
        if isinstance(widget, class_):
            getter, setter, signal = methods_
            break
    else:
        if set_value_ is None:
            raise AttributeError("I don't have any knowledge of type: '%s'" % widget)

    if get_value_:

        def get_value():
            return get_value_(widget)

    else:
        get_value = getattr(widget, getter)

    if set_value_:

        def set_value(v):
            return set_value_(widget, v)

    else:
        set_value = getattr(widget, setter)

    if conf.hasKey(key):
        try:
            v = conf.get(key)
        except TypeError:
            log.warning(
                "uistuff.loadDialogWidget: Key '%s' from conf had the wrong type '%s', ignored"
                % (key, type(conf.get(key)))
            )
            if first_value is not None:
                conf.set(key, first_value)
            else:
                conf.set(key, get_value())
        else:
            set_value(v)
    elif first_value is not None:
        conf.set(key, first_value)
        set_value(conf.get(key))
    else:
        log.warning(
            'Didn\'t load widget "%s": no conf value and no first_value arg'
            % widget_name
        )


def saveDialogWidget(widget, widget_name, config_number, get_value_=None):
    key = widget_name + "-" + str(config_number)

    if widget is None:
        raise AttributeError("key '%s' isn't in widgets" % widget_name)

    for class_, methods_ in METHODS:
        if isinstance(widget, class_):
            getter, setter, signal = methods_
            break
    else:
        if get_value_ is None:
            raise AttributeError("I don't have any knowledge of type: '%s'" % widget)

    if get_value_:

        def get_value():
            return get_value_(widget)

    else:
        get_value = getattr(widget, getter)

    if not conf.hasKey(key) or conf.get(key) != get_value():
        conf.set(key, get_value())


POSITION_NONE, POSITION_CENTER, POSITION_GOLDEN = range(3)


def keepWindowSize(key, window, defaultSize=None, defaultPosition=POSITION_NONE):
    """You should call keepWindowSize before show on your windows"""

    key = key + "window"

    def savePosition(window, *event):
        log.debug("keepWindowSize.savePosition: %s" % window.get_title())
        width = window.get_allocation().width
        height = window.get_allocation().height
        x_loc, y_loc = window.get_position()

        if width <= 0:
            log.error("Setting width = '%d' for %s to conf" % (width, key))
        if height <= 0:
            log.error("Setting height = '%d' for %s to conf" % (height, key))

        log.debug(
            "Saving window position width=%s height=%s x=%s y=%s"
            % (width, height, x_loc, y_loc)
        )
        conf.set(key + "_width", width)
        conf.set(key + "_height", height)
        conf.set(key + "_x", x_loc)
        conf.set(key + "_y", y_loc)

        return False

    window.connect("delete-event", savePosition, "delete-event")

    def loadPosition(window):
        # log.debug("keepWindowSize.loadPosition: %s" % window.title)
        # Just to make sphinx happy...
        try:
            width, height = window.get_size_request()
        except TypeError:
            pass

        if conf.hasKey(key + "_width") and conf.hasKey(key + "_height"):
            width = conf.get(key + "_width")
            height = conf.get(key + "_height")
            log.debug(f"Resizing window to width={width} height={height}")
            window.resize(width, height)

        elif defaultSize:
            width, height = defaultSize
            log.debug(f"Resizing window to width={width} height={height}")
            window.resize(width, height)

        elif key == "mainwindow":
            monitor_x, monitor_y, monitor_width, monitor_height = getMonitorBounds()
            width = int(monitor_width / 2)
            height = int(monitor_height / 4) * 3
            log.debug(f"Resizing window to width={width} height={height}")
            window.resize(width, height)

        elif key == "preferencesdialogwindow":
            monitor_x, monitor_y, monitor_width, monitor_height = getMonitorBounds()
            width = int(monitor_width / 2)
            height = int(monitor_height / 4) * 3
            window.resize(1, 1)
        else:
            monitor_x, monitor_y, monitor_width, monitor_height = getMonitorBounds()
            width = int(monitor_width / 2)
            height = int(monitor_height / 4) * 3

        if conf.hasKey(key + "_x") and conf.hasKey(key + "_y"):
            x = max(0, conf.get(key + "_x"))
            y = max(0, conf.get(key + "_y"))
            log.debug(f"Moving window to x={x} y={y}")
            window.move(x, y)

        elif defaultPosition in (POSITION_CENTER, POSITION_GOLDEN):
            monitor_x, monitor_y, monitor_width, monitor_height = getMonitorBounds()
            x_loc = int(monitor_width / 2 - width / 2) + monitor_x
            if defaultPosition == POSITION_CENTER:
                y_loc = int(monitor_height / 2 - height / 2) + monitor_y
            else:
                # Place the window on the upper golden ratio line
                y_loc = int(monitor_height / 2.618 - height / 2) + monitor_y
            log.debug(f"Moving window to x={x_loc} y={y_loc}")
            window.move(x_loc, y_loc)

    loadPosition(window)

    # In rare cases, gtk throws some gtk_size_allocation error, which is
    # probably a race condition. To avoid the window forgets its size in
    # these cases, we add this extra hook
    def callback(window):
        loadPosition(window)

    onceWhenReady(window, callback)


# Some properties can only be set, once the window is sufficiently initialized,
# This function lets you queue your request until that has happened.
def onceWhenReady(window, func, *args, **kwargs):
    def cb(window, alloc, func, *args, **kwargs):
        func(window, *args, **kwargs)
        window.disconnect(handler_id)

    handler_id = window.connect_after("size-allocate", cb, func, *args, **kwargs)


def getMonitorBounds():
    screen = Gdk.Screen.get_default()
    root_window = screen.get_root_window()
    # Just to make sphinx happy...
    try:
        ptr_window, mouse_x, mouse_y, mouse_mods = root_window.get_pointer()
        current_monitor_number = screen.get_monitor_at_point(mouse_x, mouse_y)
        monitor_geometry = screen.get_monitor_geometry(current_monitor_number)
        return (
            monitor_geometry.x,
            monitor_geometry.y,
            monitor_geometry.width,
            monitor_geometry.height,
        )
    except TypeError:
        return (0, 0, 0, 0)


def makeYellow(box):
    def on_box_expose_event(box, context):
        # box.style.paint_flat_box (box.window,
        #    Gtk.StateType.NORMAL, Gtk.ShadowType.NONE, None, box, "tooltip",
        #    box.allocation.x, box.allocation.y,
        #    box.allocation.width, box.allocation.height)
        pass

    def cb(box):
        tooltip = Gtk.Window(Gtk.WindowType.POPUP)
        tooltip.set_name("gtk-tooltip")
        tooltip.ensure_style()
        tooltipStyle = tooltip.get_style()
        box.set_style(tooltipStyle)
        box.connect("draw", on_box_expose_event)

    onceWhenReady(box, cb)


class GladeWidgets:
    """A simple class that wraps a the glade get_widget function
    into the python __getitem__ version"""

    def __init__(self, filename):
        # TODO: remove this when upstream fixes translations with Python3+Windows
        if sys.platform == "win32" and not conf.no_gettext:
            tree = ET.parse(addDataPrefix("glade/%s" % filename))
            for node in tree.iter():
                if "translatable" in node.attrib:
                    node.text = _(node.text)
                    del node.attrib["translatable"]
                if node.get("name") in ("pixbuf", "logo"):
                    node.text = addDataPrefix("glade/%s" % node.text)
            xml_text = ET.tostring(tree.getroot(), encoding="unicode", method="xml")
            self.builder = Gtk.Builder.new_from_string(xml_text, -1)
        else:
            self.builder = Gtk.Builder()
            if not conf.no_gettext:
                self.builder.set_translation_domain("pychess")
            self.builder.add_from_file(addDataPrefix("glade/%s" % filename))

    def __getitem__(self, key):
        return self.builder.get_object(key)

    def getGlade(self):
        return self.builder



import re

from sqlalchemy import select, func, and_, or_

from pychess.Utils.const import FEN_START, WHITE, BLACK, reprResult
from pychess.Database import model as dbmodel
from pychess.Database.model import (
    game,
    event,
    site,
    player,
    pl1,
    pl2,
    annotator,
    source,
    tag_game,
)


def parseDateTag(tag):
    elements = re.match(r"^([0-9\?]{4})(\.([0-9\?]{2})(\.([0-9\?]{2}))?)?$", tag)
    if elements is None:
        y, m, d = None, None, None
    else:
        elements = elements.groups()
        try:
            y = int(elements[0])
        except Exception:
            y = None
        try:
            m = int(elements[2])
        except Exception:
            m = None
        try:
            d = int(elements[4])
        except Exception:
            d = None
    return y, m, d


def save(path, model, offset, flip=False):
    game_event = model.tags["Event"]
    game_site = model.tags["Site"]

    date = model.tags["Date"]

    game_round = model.tags["Round"]

    white = repr(model.players[WHITE])
    black = repr(model.players[BLACK])

    result = model.status
    eco = model.tags["ECO"]

    time_control = model.tags["TimeControl"]
    board = int(model.tags["Board"]) if "Board" in model.tags else 0

    white_elo = model.tags["WhiteElo"]
    black_elo = model.tags["BlackElo"]

    variant = model.variant.variant

    fen = model.boards[0].board.asFen()
    fen = fen if fen != FEN_START else ""

    game_annotator = model.tags["Annotator"] if "Annotator" in model.tags else ""
    ply_count = model.ply - model.lowply

    def get_id(table, name):
        if not name:
            return None

        selection = select(table.c.id).where(table.c.name == name)
        result = conn.execute(selection)
        id_ = result.scalar()
        if id_ is None:
            result = conn.execute(table.insert().values(name=name))
            id_ = result.inserted_primary_key[0]
        return id_

    engine = dbmodel.get_engine(path)

    conn = engine.connect()
    trans = conn.begin()
    try:
        event_id = get_id(event, game_event)
        site_id = get_id(site, game_site)
        white_id = get_id(player, white)
        black_id = get_id(player, black)
        annotator_id = get_id(annotator, game_annotator)

        new_values = {
            "offset": offset,
            "offset8": (offset >> 3) << 3,
            "event_id": event_id,
            "site_id": site_id,
            "date": date,
            "round": game_round,
            "white_id": white_id,
            "black_id": black_id,
            "result": result,
            "white_elo": white_elo,
            "black_elo": black_elo,
            "ply_count": ply_count,
            "eco": eco,
            "time_control": time_control,
            "board": board,
            "fen": fen,
            "variant": variant,
            "annotator_id": annotator_id,
        }

        if hasattr(model, "game_id") and model.game_id is not None:
            result = conn.execute(
                game.update().where(game.c.id == model.game_id).values(new_values)
            )
        else:
            result = conn.execute(game.insert().values(new_values))
            model.game_id = result.inserted_primary_key[0]

        trans.commit()
    except Exception:
        trans.rollback()
        raise
    conn.close()


col2label = {
    game.c.id: "Id",
    game.c.offset: "Offset",
    game.c.offset8: "Offset8",
    pl1.c.name: "White",
    pl2.c.name: "Black",
    game.c.result: "Result",
    event.c.name: "Event",
    site.c.name: "Site",
    game.c.round: "Round",
    game.c.date: "Date",
    game.c.white_elo: "WhiteElo",
    game.c.black_elo: "BlackElo",
    game.c.ply_count: "PlyCount",
    game.c.eco: "ECO",
    game.c.time_control: "TimeControl",
    game.c.board: "Board",
    game.c.fen: "FEN",
    game.c.variant: "Variant",
    annotator.c.name: "Annotator",
}


class TagDatabase:
    def __init__(self, engine):
        self.engine = engine

        self.cols = [col.label(col2label[col]) for col in col2label]

        self.select = (
            select(*self.cols)
            .outerjoin(pl1, game.c.white_id == pl1.c.id)
            .outerjoin(pl2, game.c.black_id == pl2.c.id)
            .outerjoin(event, game.c.event_id == event.c.id)
            .outerjoin(site, game.c.site_id == site.c.id)
            .outerjoin(annotator, game.c.annotator_id == annotator.c.id)
        )

        self.select_offsets = (
            select(game.c.offset.label("Offset"))
            .outerjoin(pl1, game.c.white_id == pl1.c.id)
            .outerjoin(pl2, game.c.black_id == pl2.c.id)
            .outerjoin(event, game.c.event_id == event.c.id)
            .outerjoin(site, game.c.site_id == site.c.id)
            .outerjoin(annotator, game.c.annotator_id == annotator.c.id)
        )

        with self.engine.connect() as connection:
            self.colnames = connection.execute(self.select).keys()

        self.query = self.select
        self.order_cols = (game.c.offset, game.c.offset)
        self.is_desc = False
        self.where_tags = None
        self.where_offs = None
        self.where_offs8 = None

    @property
    def count(self):
        with self.engine.connect() as connection:
            result = connection.execute(select(func.count()).select_from(game)).scalar()
        return result

    def close(self):
        self.engine.dispose()

    def build_order_by(self, order_col, is_desc):
        self.is_desc = is_desc
        self.order_cols = (order_col, game.c.offset)

    def build_where_tags(self, tag_query):
        if tag_query is not None:
            tags = []
            if "white" in tag_query:
                if "ignore_tag_colors" in tag_query:
                    tags.append(
                        or_(
                            pl1.c.name.like("%%%s%%" % tag_query["white"]),
                            pl2.c.name.like("%%%s%%" % tag_query["white"]),
                        )
                    )
                else:
                    tags.append(pl1.c.name.like("%%%s%%" % tag_query["white"]))

            if "black" in tag_query:
                if "ignore_tag_colors" in tag_query:
                    tags.append(
                        or_(
                            pl1.c.name.like("%%%s%%" % tag_query["black"]),
                            pl2.c.name.like("%%%s%%" % tag_query["black"]),
                        )
                    )
                else:
                    tags.append(pl2.c.name.like("%%%s%%" % tag_query["black"]))

            if "event" in tag_query:
                tags.append(event.c.name.like("%%%s%%" % tag_query["event"])),

            if "site" in tag_query:
                tags.append(site.c.name.like("%%%s%%" % tag_query["site"])),

            if "eco_from" in tag_query:
                tags.append(game.c.eco >= tag_query["eco_from"])

            if "eco_to" in tag_query:
                tags.append(game.c.eco <= tag_query["eco_to"])

            if "annotator" in tag_query:
                tags.append(annotator.c.name.like("%%%s%%" % tag_query["annotator"])),

            if "variant" in tag_query:
                tags.append(game.c.variant == int(tag_query["variant"])),

            if "result" in tag_query:
                tags.append(game.c.result == reprResult.index(tag_query["result"])),

            if "date_from" in tag_query:
                tags.append(game.c.date >= tag_query["date_from"])

            if "date_to" in tag_query:
                # When date_to is not given as full date we have to prepare
                # date_to filled with some "?" to get correct query results
                # because for example "2018.??.??" is greater than "2018"
                date_to = tag_query["date_to"]
                y, m, d = parseDateTag(date_to)
                y = "%04d" % y if y is not None else "????"
                m = "%02d" % m if m is not None else "??"
                d = "%02d" % d if d is not None else "??"
                date_to = f"{y}.{m}.{d}"

                tags.append(game.c.date <= date_to)

            if "elo_from" in tag_query:
                tags.append(game.c.white_elo >= tag_query["elo_from"])
                tags.append(game.c.black_elo >= tag_query["elo_from"])

                # white_elo and black_elo are String(4) in game table
                # we need to prevent selecting games where a player elo is 999 or 400
                # when tag_query["elo_from"] is for example was set to 2200
                tags.append(game.c.white_elo < "4")
                tags.append(game.c.black_elo < "4")

            if "elo_to" in tag_query:
                tags.append(game.c.white_elo <= tag_query["elo_to"])
                tags.append(game.c.black_elo <= tag_query["elo_to"])

            self.where_tags = and_(*tags)
        else:
            self.where_tags = None

    def build_where_offs8(self, offset_list):
        if offset_list is not None and len(offset_list) > 0:
            self.where_offs8 = game.c.offset8.in_(offset_list)
        else:
            self.where_offs8 = None

    def build_where_offs(self, offset_list):
        if offset_list is not None and len(offset_list) > 0:
            self.where_offs = game.c.offset.in_(offset_list)
        else:
            self.where_offs = None

    def build_query(self):
        self.query = self.select

        if self.where_tags is not None:
            self.query = self.query.where(self.where_tags)

        if self.where_offs8 is not None:
            self.query = self.query.where(self.where_offs8)

        if self.where_offs is not None:
            self.query = self.query.where(self.where_offs)

    def get_records(self, last_seen, limit):
        self.build_query()
        # we use .where() to implement pagination because .offset() doesn't scale on big tables
        # http://sqlite.org/cvstrac/wiki?p=ScrollingCursor
        # https://stackoverflow.com/questions/21082956/sqlite-scrolling-cursor-how-to-scroll-correctly-with-duplicate-names
        if self.is_desc:
            query = (
                self.query.where(
                    or_(
                        self.order_cols[0] < last_seen[0],
                        and_(
                            self.order_cols[0] == last_seen[0],
                            self.order_cols[1] < last_seen[1],
                        ),
                    )
                )
                .order_by(self.order_cols[0].desc(), self.order_cols[1].desc())
                .limit(limit)
            )
        else:
            query = (
                self.query.where(
                    or_(
                        self.order_cols[0] > last_seen[0],
                        and_(
                            self.order_cols[0] == last_seen[0],
                            self.order_cols[1] > last_seen[1],
                        ),
                    )
                )
                .order_by(*self.order_cols)
                .limit(limit)
            )
        # with self.engine.connect() as connection:
        #    result = connection.execute(Explain(query)).fetchall(), extra={"task": "SQL"}
        # log.debug(result)

        with self.engine.connect() as connection:
            result = connection.execute(query)
            records = result.mappings().fetchall()

        return records

    def get_offsets_for_tags(self, last_seen):
        query = self.select_offsets.where(self.where_tags).where(
            game.c.offset > last_seen[1]
        )
        with self.engine.connect() as connection:
            result = connection.execute(query)
            offsets = [rec[0] for rec in result.mappings().fetchall()]
        return offsets

    def get_info(self, rec):
        where = and_(game.c.source_id == source.c.id, game.c.id == rec["Id"])
        with self.engine.connect() as connection:
            result = connection.execute(select(source.c.info).where(where)).first()
            if result is None:
                info = None
            else:
                info = result[0]
        return info

    def get_exta_tags(self, rec):
        with self.engine.connect() as connection:
            result = connection.execute(
                select(tag_game).where(tag_game.c.game_id == rec["Id"])
            )
            exta_tags = result.mappings().fetchall()
        return exta_tags



import cairo

from .__init__ import NORTH, EAST, SOUTH, WEST, CENTER
from .OverlayWindow import OverlayWindow

from math import ceil as fceil

# ceil = lambda f: int(fceil(f))


def ceil(f):
    return int(fceil(f))


class HighlightArea(OverlayWindow):
    """An entirely blue widget"""

    def __init__(self, parent):
        OverlayWindow.__init__(self, parent)
        self.cid = self.connect_after("draw", self.__onExpose)

    def showAt(self, position):
        alloc = self.myparent.get_allocation()
        if position == NORTH:
            x_loc, y_loc = 0, 0
            width, height = alloc.width, alloc.height * 0.381966011
        elif position == EAST:
            x_loc, y_loc = alloc.width * 0.618033989, 0
            width, height = alloc.width * 0.381966011, alloc.height
        elif position == SOUTH:
            x_loc, y_loc = 0, alloc.height * 0.618033989
            width, height = alloc.width, alloc.height * 0.381966011
        elif position == WEST:
            x_loc, y_loc = 0, 0
            width, height = alloc.width * 0.381966011, alloc.height
        elif position == CENTER:
            x_loc, y_loc = 0, 0
            width, height = alloc.width, alloc.height

        try:
            x_loc, y_loc = self.translateCoords(int(x_loc), int(y_loc))
            self.move(x_loc, y_loc)
        except ValueError:
            pass
            # Can't move to x,y, because top level parent has no window.

        self.resize(ceil(width), ceil(height))
        self.show()

    def __onExpose(self, self_, ctx):
        context = self.get_window().cairo_create()
        a = self_.get_allocation()
        context.rectangle(a.x, a.y, a.width, a.height)
        sc = self.get_style_context()
        found, color = sc.lookup_color("p_light_selected")
        if self.is_composited():
            context.set_operator(cairo.OPERATOR_CLEAR)
            context.set_source_rgba(0, 0, 0, 0.0)
            context.fill_preserve()
            context.set_operator(cairo.OPERATOR_OVER)
            context.set_source_rgba(color.red, color.green, color.blue, 0.5)
            context.fill()
        else:
            context.set_source_rgba(color.red, color.green, color.blue)
            context.set_operator(cairo.OPERATOR_OVER)
            context.fill()
